################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

# In contrast to sample `py_ver.py` where functions from `version.dll` are used
# to get information from version resources, this sample shows how to parse
# version resources `by hand`. This can be usefull to get string entries that
# do not use the standard names (see ctwin32.version._str_info_names). You might
# try `dump_ver_res.py %SystemRoot%\system32` to see lots of output.

import os
import sys
import ctypes
import pathlib
from types import SimpleNamespace
from pprint import pprint
from ctwin32 import (
    kernel,
    suppress_winerr,
    ERROR_ACCESS_DENIED,
    ERROR_BAD_EXE_FORMAT,
    ERROR_FILE_NOT_FOUND,
    ERROR_FILE_INVALID,
    ERROR_RESOURCE_DATA_NOT_FOUND,
    ERROR_RESOURCE_TYPE_NOT_FOUND,
    ERROR_SHARING_VIOLATION,
    LOAD_LIBRARY_AS_DATAFILE,
    LOAD_LIBRARY_AS_IMAGE_RESOURCE,
    RT_VERSION,
    )
from ctwin32.wtypes import (
    PSTR,
    PVOID,
    WORD,
    )
from ctwin32.version import VS_FIXEDFILEINFO

################################################################################

# See these pages from the `Old New Thing` (Raymond Chen) blog:
# https://devblogs.microsoft.com/oldnewthing/20061220-15/?p=28653
# https://devblogs.microsoft.com/oldnewthing/20061221-02/?p=28643
# https://devblogs.microsoft.com/oldnewthing/20061222-00/?p=28623

class VS_NODE(ctypes.Structure):
    _fields_ = (
        ("node_size", WORD),
        ("value_size", WORD),
        ("type", WORD),
        # followed by zero terminated name (WCHAR[])
        # followed by DWORD padding
        )

################################################################################

def parse_version_resource(ver_res_bytes):
    # A lot of version resources out there contain minor deviations from the
    # actual rules. This code tries to find the right balance between requiring
    # a minimum of conformance and forgiving those little glitches.

    ############################################################################

    def dword_align(x):
        return (x + 3) & (~ 3)

    ############################################################################

    def parse_ver_node(ver_res_bytes, offs):
        info = VS_NODE.from_buffer_copy(ver_res_bytes, offs)
        end = offs + info.node_size
        offs += ctypes.sizeof(VS_NODE)
        addr = ctypes.cast(PSTR(ver_res_bytes), PVOID).value + offs
        strng = ctypes.wstring_at(addr)
        return strng, dword_align(offs + (len(strng) + 1) * 2), end, info

    ############################################################################

    INVALID = ValueError("invalid version resource data")
    HEX_SET = set("abcdefABCDEF0123456789")
    name, offs, end_root, root = parse_ver_node(ver_res_bytes, 0)
    valid = (
        root.type == 0 and
        root.value_size == ctypes.sizeof(VS_FIXEDFILEINFO) and
        name == "VS_VERSION_INFO"
        )
    if not valid:
        raise INVALID

    fix = VS_FIXEDFILEINFO.from_buffer_copy(ver_res_bytes, offs)
    offs += ctypes.sizeof(VS_FIXEDFILEINFO)
    fix = {f: getattr(fix, f) for f, _ in fix._fields_}
    result = SimpleNamespace(fixed_info=fix, string_info={})
    # sizeof(VS_FIXEDFILEINFO) is a multiple of DWORDs -> no padding required

    while offs < end_root:
        name, offs, end_strings, _ = parse_ver_node(ver_res_bytes, offs)
        if name != "StringFileInfo":
            # unknown or unused (i.e. VarFileInfo) node type -> skip it
            # padding between nodes is not part of 'node_size' -> dword_align
            offs = dword_align(end_strings)
            continue

        while offs < end_strings:
            name, offs, end_translation, _ = parse_ver_node(ver_res_bytes, offs)
            if len(name) != 8 or not set(name).issubset(HEX_SET):
                raise INVALID
            translation = name[:4], name[4:]
            pairs = {}
            while offs < end_translation:
                key, offs, val_end, _ = parse_ver_node(ver_res_bytes, offs)

                # Buggy resource compilers often write the wrong value for
                # value_size. so we don't trust that and simply work with
                # what's left. But this remaining part has to have an even
                # length if we want to decode it as utf-16.
                size = (val_end - offs) & ~(0x1)
                val = ver_res_bytes[offs : offs + size]
                val = val.decode("utf-16").rstrip("\0")

                # padding between nodes is not part of 'node_size'
                offs = dword_align(val_end)
                pairs[key] = val
            result.string_info[translation] = pairs

    return result

################################################################################

EXPECTED_ERR = (
    ERROR_ACCESS_DENIED,
    ERROR_BAD_EXE_FORMAT,
    ERROR_FILE_NOT_FOUND,
    ERROR_FILE_INVALID,
    ERROR_RESOURCE_DATA_NOT_FOUND,
    ERROR_RESOURCE_TYPE_NOT_FOUND,
    ERROR_SHARING_VIOLATION,
    )
LOLI_FLAGS = LOAD_LIBRARY_AS_DATAFILE | LOAD_LIBRARY_AS_IMAGE_RESOURCE

def find_ver_res_in_file(fname):
    with suppress_winerr(*EXPECTED_ERR):
        with kernel.LoadLibraryEx(str(fname), LOLI_FLAGS) as hmod:
            for name in kernel.get_resource_names(hmod, RT_VERSION):
                addr, size = kernel.get_resource_info(hmod, name, RT_VERSION)
                yield ctypes.string_at(addr, size)

################################################################################

if __name__ == "__main__":
    start_path = os.path.expandvars(sys.argv[1]) if len(sys.argv) > 1 else "."
    p = pathlib.Path(start_path).resolve()
    for file in (e for e in p.rglob("*") if e.is_file()):
        try:
            for ver_res_bytes in find_ver_res_in_file(file):
                ver_res = parse_version_resource(ver_res_bytes)
                print(f"\n{75 * '-'}\n\n{file}\n\nfixed:")
                for name, value in ver_res.fixed_info.items():
                    print(f"    {name:18} = 0x{value:x}")
                print("\nstrings:")
                pprint(ver_res.string_info)
                print()
        except Exception as err:
            print(file, err)
            raise

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import ipaddress as _iaddr
from types import SimpleNamespace as _namespace
from collections import defaultdict as _defdict

import ctypes
from .wtypes import (
    byte_buffer,
    string_buffer,
    BYTE,
    DWORD,
    GUID,
    INT,
    PCHAR,
    PGUID,
    POINTER,
    PPVOID,
    PULONG,
    PULONGLONG,
    PVOID,
    PWSTR,
    SIZE_T,
    ULONG,
    ULONGLONG,
    WORD,
    )
from . import (
    ref,
    fun_fact,
    raise_on_err,
    ERROR_BUFFER_OVERFLOW,
    IF_TYPE_SOFTWARE_LOOPBACK,
    AF_UNSPEC,
    AF_INET,
    AF_INET6,
    )

_iph = ctypes.WinDLL("iphlpapi.dll", use_last_error=True)

################################################################################

GAA_FLAG_SKIP_UNICAST                = 0x0001
GAA_FLAG_SKIP_ANYCAST                = 0x0002
GAA_FLAG_SKIP_MULTICAST              = 0x0004
GAA_FLAG_SKIP_DNS_SERVER             = 0x0008
GAA_FLAG_INCLUDE_PREFIX              = 0x0010
GAA_FLAG_SKIP_FRIENDLY_NAME          = 0x0020
GAA_FLAG_INCLUDE_WINS_INFO           = 0x0040
GAA_FLAG_INCLUDE_GATEWAYS            = 0x0080
GAA_FLAG_INCLUDE_ALL_INTERFACES      = 0x0100
GAA_FLAG_INCLUDE_ALL_COMPARTMENTS    = 0x0200
GAA_FLAG_INCLUDE_TUNNEL_BINDINGORDER = 0x0400
GAA_FLAG_SKIP_DNS_INFO               = 0x0800

################################################################################

class _STRUCTURE(ctypes.Structure):
    _fields_ = (("Length", ULONG), ("Flags", ULONG))

class _UNION(ctypes.Union):
    _fields_ = (("Alignment", ULONGLONG), ("s", _STRUCTURE))
    _anonymous_ = ("s",)

class SOCKADDR(ctypes.Structure):
    _fields_ = (("sa_family", WORD), ("sa_data", BYTE * 14))

PSOCKADDR = POINTER(SOCKADDR)

################################################################################

class SOCKET_ADDRESS(ctypes.Structure):
    _fields_ = (
        ("lpSockaddr", PSOCKADDR),
        ("iSockaddrLength", INT)
        )
PSOCKET_ADDRESS = POINTER(SOCKET_ADDRESS)

################################################################################

class IP_ADAPTER_UNICAST_ADDRESS(ctypes.Structure):
    pass
PIP_ADAPTER_UNICAST_ADDRESS = POINTER(IP_ADAPTER_UNICAST_ADDRESS)

IP_ADAPTER_UNICAST_ADDRESS._fields_ = (
    ("u", _UNION),
    ("Next", PIP_ADAPTER_UNICAST_ADDRESS),
    ("Address", SOCKET_ADDRESS),
    # we do not need any field after 'Address', so we do not define them
    )

################################################################################

class IP_ADAPTER_PREFIX(ctypes.Structure):
    pass
PIP_ADAPTER_PREFIX = POINTER(IP_ADAPTER_PREFIX)
IP_ADAPTER_PREFIX._fields_ = (
    ("u", _UNION),
    ("Next", PIP_ADAPTER_PREFIX),
    ("Address", SOCKET_ADDRESS),
    ("PrefixLength", ULONG)
    )

################################################################################

class IP_ADAPTER_ADDRESSES(ctypes.Structure):
    pass
PIP_ADAPTER_ADDRESSES = POINTER(IP_ADAPTER_ADDRESSES)
IP_ADAPTER_ADDRESSES._fields_ = (
    ("u", _UNION),
    ("Next", PIP_ADAPTER_ADDRESSES),
    ("AdapterName", PCHAR),
    ("FirstUnicastAddress", PIP_ADAPTER_UNICAST_ADDRESS),
    ("FirstAnycastAddress", PVOID),
    ("FirstMulticastAddress", PVOID),
    ("FirstDnsServerAddress", PVOID),
    ("DnsSuffix", PWSTR),
    ("Description", PWSTR),
    ("FriendlyName", PWSTR),
    ("PhysicalAddress", BYTE * 8),
    ("PhysicalAddressLength", ULONG),
    ("Flags", ULONG),
    ("Mtu", ULONG),
    ("IfType", ULONG),
    ("OperStatus", INT),
    ("Ipv6IfIndex", ULONG),
    ("ZoneIndices", ULONG * 16),
    ("FirstPrefix", PIP_ADAPTER_PREFIX)
    # we do not need any field after 'FirstPrefix',
    # so we do not define them
    )

################################################################################

class S_UN_B(ctypes.Structure):
    _fields_ = (
        ("s_b1", BYTE),
        ("s_b2", BYTE),
        ("s_b3", BYTE),
        ("s_b4", BYTE)
        )

class S_UN_W(ctypes.Structure):
    _fields_ = (("s_w1", WORD), ("s_w2", WORD))

class S_UN(ctypes.Union):
    _fields_ = (
        ("S_un_b", S_UN_B),
        ("S_un_w", S_UN_W),
        ("S_addr", ULONG.__ctype_be__)
        )

class IN_ADDR(ctypes.Structure):
    _fields_ = (("S_un", S_UN),)
PIN_ADDR = POINTER(IN_ADDR)

################################################################################

class SOCKADDR_IN(ctypes.Structure):
    _fields_ = (
        ("sin_family", WORD),
        ("sin_port", WORD),
        ("sin_addr", IN_ADDR),
        ("sin_zero", BYTE * 8)
        )
PSOCKADDR_IN = POINTER(SOCKADDR_IN)

################################################################################

class IN6_ADDR(ctypes.Union):
    _fields_ = (("Byte", BYTE * 16), ("Word", WORD * 8))
PIN6_ADDR = POINTER(IN6_ADDR)

################################################################################

class SOCKADDR_IN6(ctypes.Structure):
    _fields_ = (
        ("sin6_family", WORD),
        ("sin6_port", WORD),
        ("sin6_flowinfo", ULONG),
        ("sin6_addr", IN6_ADDR),
        ("sin6_scope_id", ULONG),
        )
PSOCKADDR_IN6 = POINTER(SOCKADDR_IN6)

################################################################################

class SOCKADDR_INET(ctypes.Union):
    _fields_ = (
        ("Ipv4", SOCKADDR_IN),
        ("Ipv6", SOCKADDR_IN6),
        ("si_family", WORD),
        )

    def get_ipaddr(self):
        fam = self.Ipv4.sin_family
        if fam == AF_INET:
            return _iaddr.IPv4Address(self.Ipv4.sin_addr.S_un.S_addr)
        elif fam == AF_INET6:
            return _iaddr.IPv6Address(bytes(self.Ipv6.sin6_addr.Byte))
        else:
            raise ValueError(f"unsupported address family: {fam}")

################################################################################

def _sock_addr_to_ip_addr(p_sock_addr):
    fam = p_sock_addr.contents.sa_family
    if fam == AF_INET:
        addr = ctypes.cast(p_sock_addr, PSOCKADDR_IN).contents
        return _iaddr.IPv4Address(addr.sin_addr.S_un.S_addr)
    elif fam == AF_INET6:
        addr = ctypes.cast(p_sock_addr, PSOCKADDR_IN6).contents
        ip = _iaddr.IPv6Address(bytes(addr.sin6_addr.Byte))
        if addr.sin6_scope_id:
            ip = _iaddr.IPv6Address(f"{ip}%{addr.sin6_scope_id}")
        return ip
    else:
        raise ValueError(f"unsupported address family: {fam}")

################################################################################

_GetAdaptersAddresses = fun_fact(
    _iph.GetAdaptersAddresses, (
        ULONG,
        ULONG,
        ULONG,
        PVOID,
        PIP_ADAPTER_ADDRESSES,
        PULONG
        )
    )

################################################################################

def _best_prefix_len(ip, prefixes):
    best_len = 0
    for pfx, plen in prefixes:
        skip = (
            ip.version != pfx.version or
            best_len > plen or
            (best_len and plen == ip.max_prefixlen)
            )
        if not skip:
            mask = -1 << (ip.max_prefixlen - plen)
            if int(ip) & mask == int(pfx) & mask:
                best_len = plen

    return best_len if best_len else ip.max_prefixlen

################################################################################

def _adapter_addresses_to_interfaces(p_adresses, include_loopback):
    result = _defdict(list)
    while p_adresses:
        adptr_addr = p_adresses.contents
        p_adresses = adptr_addr.Next
        not_loopback = adptr_addr.IfType != IF_TYPE_SOFTWARE_LOOPBACK
        if not_loopback or include_loopback:
            pfx_ptr = adptr_addr.FirstPrefix
            prefixes = []
            while pfx_ptr:
                pfx = pfx_ptr.contents
                pfx_ptr = pfx.Next
                prefix = _sock_addr_to_ip_addr(pfx.Address.lpSockaddr)
                prefixes.append((prefix, pfx.PrefixLength))
            adptr_name = ctypes.string_at(adptr_addr.AdapterName).decode()
            pfua = adptr_addr.FirstUnicastAddress
            while pfua:
                fua = pfua.contents
                pfua = fua.Next
                ip = _sock_addr_to_ip_addr(fua.Address.lpSockaddr)
                plen = _best_prefix_len(ip, prefixes)
                result[adptr_name].append(_iaddr.ip_interface(f"{ip}/{plen}"))

    return dict(result)  # no more default values

################################################################################

def _ver_to_fam(ver):
    return AF_INET if ver == 4 else (AF_INET6 if ver == 6 else AF_UNSPEC)

################################################################################

def get_host_interfaces(version=4, include_loopback=False):
    "returns the list of the ip interfaces of the local network adapters"
    fam = _ver_to_fam(version)
    flags = (
        GAA_FLAG_INCLUDE_PREFIX |
        GAA_FLAG_SKIP_ANYCAST |
        GAA_FLAG_SKIP_MULTICAST
        )
    blen = ULONG(16 * 1024)
    error = ERROR_BUFFER_OVERFLOW
    while error == ERROR_BUFFER_OVERFLOW:
        buffer = byte_buffer(blen.value)
        p_addr = ctypes.cast(buffer, PIP_ADAPTER_ADDRESSES)
        error = _GetAdaptersAddresses(fam, flags, None, p_addr, ref(blen))
    raise_on_err(error)

    return _adapter_addresses_to_interfaces(p_addr, include_loopback)

################################################################################

IF_MAX_PHYS_ADDRESS_LENGTH = 32

class MIB_IPNET_ROW2(ctypes.Structure):
    _fields_ = (
        ("Address", SOCKADDR_INET),
        ("InterfaceIndex", ULONG),
        ("InterfaceLuid", ULONGLONG),
        ("PhysicalAddress", BYTE * IF_MAX_PHYS_ADDRESS_LENGTH),
        ("PhysicalAddressLength", DWORD),
        ("State", DWORD),
        ("Flags", BYTE),
        ("ReachabilityTime", ULONG),
        )

################################################################################

FreeMibTable = fun_fact(_iph.FreeMibTable, (None, PVOID))

################################################################################

_GetIpNetTable2 = fun_fact(_iph.GetIpNetTable2, (DWORD, WORD, PPVOID))

def GetIpNetTable2(version=0):
    ptr = PVOID()
    raise_on_err(_GetIpNetTable2(_ver_to_fam(version), ref(ptr)))
    try:
        num = ctypes.cast(ptr, PULONG).contents.value

        class MIB_IPNETTABLE2(ctypes.Structure):
            _fields_ = (
                ("NumEntries", ULONG),
                ("Table", MIB_IPNET_ROW2 * num),
                )

        return [
            _namespace(
                index=e.InterfaceIndex,
                luid=e.InterfaceLuid,
                if_type=e.InterfaceLuid >> 48,
                phys_addr=e.PhysicalAddress[:e.PhysicalAddressLength],
                addr=e.Address.get_ipaddr(),
                state=e.State,
                flags=e.Flags,
                reach_time=e.ReachabilityTime
                )
            for e in MIB_IPNETTABLE2.from_address(ptr.value).Table
            ]

    finally:
        FreeMibTable(ptr)

################################################################################

_ConvertInterfaceGuidToLuid = fun_fact(
    _iph.ConvertInterfaceGuidToLuid,
    (DWORD, PGUID, PULONGLONG)
    )

def ConvertInterfaceGuidToLuid(guid):
    guid = GUID(guid)
    luid = ULONGLONG()
    raise_on_err(_ConvertInterfaceGuidToLuid(ref(guid), ref(luid)))
    return luid.value

################################################################################

_ConvertInterfaceIndexToLuid = fun_fact(
    _iph.ConvertInterfaceIndexToLuid,
    (DWORD, ULONG, PULONGLONG)
    )

def ConvertInterfaceIndexToLuid(idx):
    luid = ULONGLONG()
    raise_on_err(_ConvertInterfaceIndexToLuid(idx, ref(luid)))
    return luid.value

################################################################################

_ConvertInterfaceLuidToAlias = fun_fact(
    _iph.ConvertInterfaceLuidToAlias,
    (DWORD, PULONGLONG, PWSTR, SIZE_T)
    )

IF_MAX_STRING_SIZE = 256

def ConvertInterfaceLuidToAlias(luid):
    luid = ULONGLONG(luid)
    size = SIZE_T(IF_MAX_STRING_SIZE + 1)
    alias = string_buffer(size.value)
    raise_on_err(_ConvertInterfaceLuidToAlias(ref(luid), alias, size))
    return alias.value

################################################################################

_ConvertInterfaceLuidToName = fun_fact(
    _iph.ConvertInterfaceLuidToNameW,
    (DWORD, PULONGLONG, PWSTR, SIZE_T)
    )

def ConvertInterfaceLuidToName(luid):
    luid = ULONGLONG(luid)
    size = SIZE_T(IF_MAX_STRING_SIZE + 1)
    name = string_buffer(size.value)
    raise_on_err(_ConvertInterfaceLuidToName(ref(luid), name, size))
    return name.value

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################
#
# This sample shows how to find files with hardlinks - or to be more precise -
# files with more than one hardlink.
#
################################################################################

import os
import uuid
from pathlib import Path
from ctwin32 import (
    kernel,
    suppress_winerr,
    ERROR_SHARING_VIOLATION,
    GENERIC_READ,
    ERROR_ACCESS_DENIED,
    )

################################################################################

def walk_files(pth):
    for root, _, files in pth.walk():
        for f in files:
            yield root / f

################################################################################

def get_file_id(name):
    with suppress_winerr(ERROR_ACCESS_DENIED):
        with kernel.create_file(str(name), GENERIC_READ) as hdl:
            info = kernel.GetFileInformationByHandle(hdl)
        return (info.nFileIndexHigh << 32) | info.nFileIndexLow
    return int.from_bytes(uuid.uuid4().bytes)

################################################################################

def hardlinks(name):
    drive = Path(name).resolve().drive
    try:
        hl = [f"{drive}{n}" for n in kernel.find_all_filenames(str(name))]
    except OSError as e:
        print(e, name)
        hl = [str(name)]
    fid = 0 if len(hl) < 2 else get_file_id(name)
    return fid, sorted(hl, key=len)

################################################################################

if __name__ == "__main__":

    seen = set()
    for name in walk_files(Path(os.environ["SYSTEMDRIVE"] + "\\")):
        with suppress_winerr(ERROR_SHARING_VIOLATION):
            fid, h = hardlinks(name)
            if len(h) > 1 and fid not in seen:
                seen.add(fid)
                print(h[0])
                for n in h[1:]:
                    print(f"    {n}")
                print()

    print(f"{len(seen)} files with more than one hardlink")

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import sys
import ctypes
from uuid import UUID
from datetime import datetime

# avoid circular dependency: ctwin32 -> wtypes -> ctwin32
MAX_PATH = 260

################################################################################

# integral types

ENDIANNESS = sys.byteorder

BYTE = ctypes.c_ubyte
CHAR = ctypes.c_char
BOOLEAN = BYTE

WCHAR = ctypes.c_wchar
WCHAR_SIZE = ctypes.sizeof(WCHAR)
SHORT = ctypes.c_short
WORD = USHORT = ctypes.c_ushort

INT = ctypes.c_int
UINT = ctypes.c_uint
LONG = BOOL = HRESULT = NTSTATUS = ctypes.c_long
DWORD = ULONG = ctypes.c_ulong

# While the large integers are defined as a struct of low and high part, they
# are also unions containing a long long part. That is why we can simply define
# them as long longs without changing the aligment. For FILETIME (see below)
# this is not the case.

LARGE_INTEGER = LONGLONG = ctypes.c_longlong
ULARGE_INTEGER = ULONGLONG = ctypes.c_ulonglong

UINT_PTR = WPARAM = SIZE_T = ULONG_PTR = ctypes.c_size_t
INT_PTR = LPARAM = SSIZE_T = LRESULT = LONG_PTR = ctypes.c_ssize_t

################################################################################

# floating point types

FLOAT = ctypes.c_float
DOUBLE = ctypes.c_double

################################################################################

# handle types

HANDLE = HINSTANCE = HMODULE = HWND = ctypes.c_void_p

################################################################################

# byte and string buffers

def byte_buffer(init, size=None):
    if isinstance(init, bytes):
        if size is None:
            size = 1 + len(init)
        (buf := (CHAR * size)()).value = init
        return buf
    elif isinstance(init, int):
        return (CHAR * init)()
    raise TypeError(init)

def wchar_len_sz(wstr):
    return 1 + sum(2 if ord(c) > 0xFFFF else 1 for c in wstr)

def string_buffer(init, size=None):
    if isinstance(init, str):
        if size is None:
            size = wchar_len_sz(init)
        (buf := (WCHAR * size)()).value = init
        return buf
    elif isinstance(init, int):
        return (WCHAR * init)()
    raise TypeError(init)

################################################################################

# error handling based on ctypes error shadow copy

def WinError(code=None, descr=None):
    if code is None:
        code = ctypes.get_last_error()
    if descr is None:
        descr = ctypes.FormatError(code).strip()
    return OSError(None, descr, None, code)

################################################################################

# some structure definitions

class GUID(ULONG * 4):         # using ULONG for correct alignment
    def __init__(self, init=None):
        # init is None or str or GUID or UUID or something that can be
        # converted to bytes (like comtypes.GUID). If init is None we simply
        # keep the initial state of 'all bytes are zero'.
        if init is not None:
            if isinstance(init, str):
                src = UUID(init).bytes_le
            elif isinstance(init, GUID):
                src = init
            elif isinstance(init, UUID):
                src = init.bytes_le
            else:
                src = bytes(init)
            ctypes.memmove(self, src, ctypes.sizeof(self))

    def uuid(self):
        return UUID(bytes_le=bytes(self))

    def __str__(self):
        return f"{{{self.uuid()!s}}}"

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.uuid()!s}')"

    def __int__(self):
        return int.from_bytes(bytes(self), ENDIANNESS)

    def __eq__(self, other):
        try:
            return bytes(self) == bytes(other)
        except Exception:   # noqa: BLE001
            return False

    def __hash__(self):
        return int(self)

################################################################################

class FILETIME(ctypes.Structure):
    "Time in 100 nanosecond steps since January 1, 1601 (UTC)"
    # cannot represent FILETIME as ctypes.c_ulonglong since that would change
    # the alignment
    _fields_ = (
        ("dwLowDateTime", DWORD),
        ("dwHighDateTime", DWORD),
        )

    def __init__(self, i64=0):
        self.dwLowDateTime = i64 & 0xffffffff
        self.dwHighDateTime = i64 >> 32

    def __int__(self):
        return self.dwLowDateTime | (self.dwHighDateTime << 32)

    def __add__(self, other):
        return self.__class__(int(self) + int(other))

    def __sub__(self, other):
        return self.__class__(int(self) - int(other))

    def __iadd__(self, other):
        i64 = int(self) + int(other)
        self.dwLowDateTime = i64 & 0xffffffff
        self.dwHighDateTime = i64 >> 32
        return self

    def __isub__(self, other):
        i64 = int(self) - int(other)
        self.dwLowDateTime = i64 & 0xffffffff
        self.dwHighDateTime = i64 >> 32
        return self

    def __repr__(self):
        return f"{self.__class__.__name__}({int(self)})"

################################################################################

class SYSTEMTIME(ctypes.Structure):
    _fields_ = (
        ("wYear",         WORD),
        ("wMonth",        WORD),
        ("wDayOfWeek",    WORD),
        ("wDay",          WORD),
        ("wHour",         WORD),
        ("wMinute",       WORD),
        ("wSecond",       WORD),
        ("wMilliseconds", WORD)
        )

    ############################################################################

    def to_datetime(self, tzinfo=None):
        return datetime(
            self.wYear,
            self.wMonth,
            self.wDay,
            self.wHour,
            self.wMinute,
            self.wSecond,
            self.wMilliseconds * 1000,
            tzinfo
            )

    ############################################################################

    def from_datetime(self, dt):
        self.wYear = dt.year
        self.wMonth = dt.month
        self.wDay = dt.day
        self.wHour = dt.hour
        self.wMinute = dt.minute
        self.wSecond = dt.second
        self.wMilliseconds = dt.microsecond // 1000
        dow = dt.isoweekday()
        self.wDayOfWeek = 0 if dow == 7 else dow
        return self

    ############################################################################

    def to_struct_time(self):
        return self.to_datetime().timetuple()

    ############################################################################

    def from_struct_time(self, st):
        self.wYear = st.tm_year
        self.wMonth = st.tm_mon
        self.wDay = st.tm_mday
        self.wHour = st.tm_hour
        self.wMinute = st.tm_min
        self.wSecond = st.tm_sec
        self.wMilliseconds = 0
        self.wDayOfWeek = 0 if st.tm_wday == 6 else st.tm_wday + 1
        return self

    ############################################################################

    def __repr__(self):
        flds = ", ".join([f"{n}={getattr(self, n)}" for n, _ in self._fields_])
        return f"{self.__class__.__name__}({flds})"

################################################################################

class POINT(ctypes.Structure):
    _fields_ = (
        ("x", LONG),
        ("y", LONG)
        )

    @classmethod
    def from_lparam(cls, lp):
        return cls(lp & 0xffff, (lp >> 16) & 0xffff)

    def as_lparam(self):
        return (self.x & 0xffff) | ((self.y & 0xffff) << 16)

    def copy(self):
        return self.__class__(self.x, self.y)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.x}, {self.y})"

################################################################################

class RECT(ctypes.Structure):
    _fields_ = (
        ("left", LONG),
        ("top", LONG),
        ("right", LONG),
        ("bottom", LONG)
        )

    @property
    def width(self):
        return self.right - self.left

    @property
    def height(self):
        return self.bottom - self.top

    @property
    def center(self):
        return (self.left + self.right) // 2, (self.top + self.bottom) // 2

    def copy(self):
        return self.__class__(self.left, self.top, self.right, self.bottom)

    def __repr__(self):
        name = self.__class__.__name__
        return f"{name}({self.left}, {self.top}, {self.right}, {self.bottom})"

################################################################################

class LUID(ctypes.Structure):
    _fields_ = (
        ("LowPart", DWORD),
        ("HighPart", LONG)
        )

    def __init__(self, value=None):
        if value is None:
            value = 0
        self.LowPart = value & 0xffffffff
        self.HighPart = value >> 32

    def __int__(self):
        return self.LowPart | (self.HighPart << 32)

    def __repr__(self):
        return f"{self.__class__.__name__}({int(self)})"

################################################################################

class CallbackContext(ctypes.Structure):
    _fields_ = (
        ("callback", ctypes.py_object),
        ("context", ctypes.py_object)
        )

################################################################################

class LOGFONT(ctypes.Structure):
    _fields_ = (
        ("lfHeight", LONG),
        ("lfWidth", LONG),
        ("lfEscapement", LONG),
        ("lfOrientation", LONG),
        ("lfWeight", LONG),
        ("lfItalic", BYTE),
        ("lfUnderline", BYTE),
        ("lfStrikeOut", BYTE),
        ("lfCharSet", BYTE),
        ("lfOutPrecision", BYTE),
        ("lfClipPrecision", BYTE),
        ("lfQuality", BYTE),
        ("lfPitchAndFamily", BYTE),
        ("lfFaceName", WCHAR * 32),
        )

################################################################################

class OSVERSIONINFOEX(ctypes.Structure):
    _fields_ = (
        ("dwOSVersionInfoSize", DWORD),
        ("dwMajorVersion", DWORD),
        ("dwMinorVersion", DWORD),
        ("dwBuildNumber", DWORD),
        ("dwPlatformId", DWORD),
        ("szCSDVersion", WCHAR * 128),
        ("wServicePackMajor", WORD),
        ("wServicePackMinor", WORD),
        ("wSuiteMask", WORD),
        ("wProductType", BYTE),
        ("wReserved", BYTE),
        )

    def __init__(self):
        self.dwOSVersionInfoSize = ctypes.sizeof(self)

################################################################################

class WIN32_FIND_DATA(ctypes.Structure):
    _fields_ = (
        ("dwFileAttributes", DWORD),
        ("ftCreationTime", FILETIME),
        ("ftLastAccessTime", FILETIME),
        ("ftLastWriteTime", FILETIME),
        ("nFileSizeHigh", DWORD),
        ("nFileSizeLow", DWORD),
        ("dwReserved0", DWORD),
        ("dwReserved1", DWORD),
        ("cFileName", WCHAR * MAX_PATH),
        ("cAlternateFileName", WCHAR * 14),
        ("dwFileType", DWORD),     # obsolete - do not use
        ("dwCreatorType", DWORD),  # obsolete - do not use
        ("wFinderFlags", WORD),    # obsolete - do not use
        )

################################################################################

# pointer types

PTR_HAS_32_BITS = ctypes.sizeof(ctypes.c_void_p) == 4
PTR_HAS_64_BITS = ctypes.sizeof(ctypes.c_void_p) == 8

PWSTR = ctypes.c_wchar_p
PSTR = ctypes.c_char_p
PVOID = ctypes.c_void_p
POINTER = ctypes.POINTER
PPWSTR = POINTER(PWSTR)
PPVOID = POINTER(PVOID)
PBYTE = POINTER(BYTE)
PCHAR = POINTER(CHAR)
PBOOLEAN = POINTER(BOOLEAN)
PWCHAR = POINTER(WCHAR)
PSHORT = POINTER(SHORT)
PUSHORT = PWORD = POINTER(USHORT)
PINT = POINTER(INT)
PUINT = POINTER(UINT)
PLONG = PBOOL = POINTER(LONG)
PULONG = PDWORD = POINTER(ULONG)
PUINT_PTR = PWPARAM = PSIZE_T = PULONG_PTR = POINTER(UINT_PTR)
PLARGE_INTEGER = POINTER(LARGE_INTEGER)
PULARGE_INTEGER = POINTER(ULARGE_INTEGER)
PHANDLE = POINTER(HANDLE)
PGUID = POINTER(GUID)
PFILETIME = POINTER(FILETIME)
PSYSTEMTIME = POINTER(SYSTEMTIME)
PPOINT = POINTER(POINT)
PRECT = POINTER(RECT)
PULONGLONG = POINTER(ULONGLONG)
PLUID = POINTER(LUID)
PPLUID = POINTER(PLUID)
CallbackContextPtr = POINTER(CallbackContext)
PLOGFONT = POINTER(LOGFONT)
POSVERSIONINFOEX = POINTER(OSVERSIONINFOEX)
PWIN32_FIND_DATA = POINTER(WIN32_FIND_DATA)

def ptr_addr(ct_obj):
    return PVOID.from_buffer(ct_obj).value

################################################################################

# A class that allows to create types (by multiple inheritance) that are based
# on ctypes._SimpleCData and can be used as context managers (i.e. can be used
# in 'with' statements) in order to ensure that handles are orderly closed.
# When defining such a type it must be configured with the approriate base
# class and two parameters:
# - close_func: A function to be called on an instance to close it.
# - invalid: The value that represents an invalid object.
#
# e.g.:
# class KHNDL(
#   ScdToBeClosed,
#   HANDLE,
#   close_func=CloseHandle,
#   invalid=0
#   ):
#
# Since most of the time HANDLE/ctypes.c_void_p will be the _SimpleCData base
# class, we have to deal with the fact that ctypes.c_void_p will report 'None'
# for its attribute 'value', if the actual integer value is zero.

class ScdToBeClosed():

    def __init_subclass__(cls, /, close_func, invalid, **kwargs):
        if not issubclass(cls, ctypes._SimpleCData):
            raise TypeError("must inherit from ctypes._SimpleCData")
        super().__init_subclass__(**kwargs)
        cls.close_func = close_func
        cls.invalid_value = int(0 if invalid is None else invalid)

    def __init__(self, init=None):
        if hasattr(init, "value"):
            init = 0 if init.value is None else init.value
        if init is None:
            init = self.invalid_value
        self.value = init

    def close(self):
        if self.is_valid():
            self.close_func()
            self.value = self.invalid_value

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @classmethod
    def from_param(cls, obj):
        return obj if isinstance(obj, cls) else cls(obj)

    def __int__(self):
        return 0 if self.value is None else self.value

    def is_valid(self):
        return int(self) != self.invalid_value

    def raise_on_invalid(self):
        if int(self) == self.invalid_value:
            raise WinError()

################################################################################

# c_void_p has a flaw: neither can c_void_p.from_param() process a py_object,
# nor does cast() allow to convert a py_object to a c_void_p
# (cast(py_object(obj), c_void_p) fails). Therefore we need this odd way of
# converting a python object pointer to c_void_p.

def pvoid_from_obj(obj):
    return PVOID.from_buffer(ctypes.py_object(obj))

# On the other hand: As long as CPython keeps using the object address as the
# return value for `id()`, you can simply use `id(obj)` instead of
# `pvoid_from_obj(obj)`.

################################################################################

class ArgcArgvFromArgs():
    """Converts a list of strings into a buffer that contains the strings
    and an array of pointers to these strings. The number of strings and the
    address of the pointer array are provided by the attributes `argc` and
    `argv`.
    """
    def __init__(self, args):
        if not args:
            self._argc = 0
            self._argv = None
        else:
            self._argc = len(args)
            chain = "\0".join(args) + "\0"

            class ArgumentBuffer(ctypes.Structure):
                _fields_ = (
                    ("pointers", PWSTR * self._argc),
                    ("strings", WCHAR * (wchar_len_sz(chain) - 1)),
                    )

            self._buffer = ArgumentBuffer(strings=chain)
            self._argv = ctypes.addressof(self._buffer)
            str_addr = self._argv + ArgumentBuffer.strings.offset
            for i, arg in enumerate(args):
                self._buffer.pointers[i] = str_addr
                str_addr += WCHAR_SIZE * wchar_len_sz(arg)

    @property
    def argc(self):
        return self._argc

    @property
    def argv(self):
        return self._argv

################################################################################

class UNICODE_STRING(ctypes.Structure):
    _fields_ = (
        ("Length", WORD),
        ("MaximumLength", WORD),
        ("Buffer", PWSTR),
        )

    def __str__(self):
        return ctypes.wstring_at(
            self.Buffer,
            self.Length // WCHAR_SIZE
            )

PUNICODE_STRING = POINTER(UNICODE_STRING)

################################################################################

def UnicodeStrBuffer(init):
    if isinstance(init, str):
        ws = wchar_len_sz(init)
    else:
        ws = init
        init = "\0" * init

    class SELF_CONTAINED_US(ctypes.Structure):
        _fields_ = (
            ("us", UNICODE_STRING),
            ("buf", WCHAR * ws),
            )

        def __init__(self, init):
            li = ws * WCHAR_SIZE
            baddr = ctypes.addressof(self) + __class__.buf.offset
            super().__init__((li - WCHAR_SIZE, li, baddr), init)

        @property
        def ptr(self):
            return PUNICODE_STRING(self.us)

        @property
        def str(self):
            return str(self.us)

    return SELF_CONTAINED_US(init)

################################################################################

def UnicodeStrArray(strings):
    num_strings = len(strings)
    chain = "\0".join(strings)
    buf_len = wchar_len_sz(chain)

    class SELF_CONTAINED_USA(ctypes.Structure):
        _fields_ = (
            ("us", UNICODE_STRING * num_strings),
            ("buf", WCHAR * buf_len),
            )

        def __init__(self, strings, chain):
            super().__init__(buf=chain)
            addr = ctypes.addressof(self) + __class__.buf.offset
            for i, s in enumerate(strings):
                max_len = wchar_len_sz(s) * WCHAR_SIZE
                self.us[i].Length = max_len - WCHAR_SIZE
                self.us[i].MaximumLength = max_len
                self.us[i].Buffer = addr
                addr += max_len

        @property
        def ptr(self):
            return PUNICODE_STRING(self.us)

    return SELF_CONTAINED_USA(strings, chain)

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import sys
import argparse
from ctwin32 import (
    ctypes,
    user,
    advapi,
    shell,
    wtypes,
    HWND_BROADCAST,
    WM_SETTINGCHANGE,
    SMTO_NORMAL,
    REG_SZ,
    REG_EXPAND_SZ,
    KEY_READ,
    KEY_WRITE,
    )

################################################################################

def env_var_root(system=False, access=KEY_READ):
    if system:
        pth = r"SYSTEM\CurrentControlSet\Control\Session Manager"
        return advapi.RegOpenKeyEx(advapi.HKLM, pth, access)
    else:
        return advapi.HKCU

################################################################################

def env_var_key(root, access=KEY_READ):
    return advapi.RegOpenKeyEx(root, "Environment", access)

################################################################################

def is_persistent_env_var(name, system=False):
    with env_var_root(system) as root:
        with env_var_key(root) as key:
            result = False
            try:
                val, typ = advapi.RegQueryValueEx(key, name)
                result = (typ in (REG_SZ, REG_EXPAND_SZ)) and bool(val)
            except OSError:
                pass
            return result

################################################################################

def broadcast_env_change():
    estr = wtypes.string_buffer("Environment")
    user.SendMessageTimeout(
        HWND_BROADCAST,
        WM_SETTINGCHANGE,
        0,
        ctypes.addressof(estr),
        SMTO_NORMAL,
        500
        )

################################################################################

def persist_env_var(name, value, system=False, do_broadcast=False):
    access = KEY_WRITE | KEY_READ
    with env_var_root(system, access) as root:
        with env_var_key(root, access) as key:
            if not value:
                advapi.RegDeleteValue(key, name)
            else:
                advapi.reg_set_str(key, name, value)
        if do_broadcast:
            broadcast_env_change()

################################################################################

def persist_user_env_block(nv_dict, system=False):
    for n, v in nv_dict.items():
        persist_env_var(n, v, system, False)
    broadcast_env_change()

################################################################################

def get_env_block(system=False):
    with env_var_root(system) as root:
        with env_var_key(root, KEY_READ) as key:
            return {
                nam: val for nam, val, typ in advapi.reg_enum_values(key)
                if typ in (REG_SZ, REG_EXPAND_SZ)
                }

################################################################################

def parse_args():
    ape = argparse.ArgumentParser(
        description="set environment variables persistently (like setx)"
        )
    ape.add_argument(
        "-s",
        "--system",
        action="store_true",
        help="set system variable (as opposed to user variable)"
        )
    ape.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="print final variables"
        )
    ape.add_argument("name", help="name of variable")
    ape.add_argument(
        "value",
        help="value of variable (omitting it will delete the variable)",
        nargs="?",
        default="",
        )
    return ape.parse_args()

################################################################################

def main(name, value, system, verbose):
    persist_env_var(name, value, system, True)
    if verbose:
        print(f"variables for {'system' if args.system else 'user'}:")
        for k, v in get_env_block(args.system).items():
            print(f"    {k} = {v}")

################################################################################

if __name__ == "__main__":
    args = parse_args()

    # Setting system variables requires administrative privileges.
    if args.system and not advapi.running_as_admin():
        shell.elevate(sys.executable, *sys.argv)
    else:
        main(args.name, args.value, args.system, args.verbose)

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

# inspired by
# https://docs.microsoft.com/en-us/sysinternals/downloads/logonsessions

import sys
from ctwin32.advapi import running_as_admin
from ctwin32.secur import LsaEnumerateLogonSessions

if __name__ == "__main__":
    if not running_as_admin():
        m = "Need to run as administrator in order to enumerate logon sessions."
        print(m, file=sys.stderr)
        sys.exit(1)

    verbose = "-verbose" in sys.argv
    lotypes = [
        "(none)",
        "???UNDEFINED???",
        "Interactive",
        "Network",
        "Batch",
        "Service",
        "Proxy",
        "Unlock",
        "NetworkCleartext",
        "NewCredentials",
        "RemoteInteractive",
        "CachedInteractive",
        "CachedRemoteInteractive",
        "CachedUnlock"
        ]

    def fmtt(dt):
        return dt.isoformat(" ", "seconds")

    for idx, ses in enumerate(LsaEnumerateLogonSessions()):
        lot = (
            lotypes[ses.LogonType] if 0 <= ses.LogonType < len(lotypes)
            else f"{ses.LogonType} (unknown)"
            )
        print(f"\n[{idx}] Logon session {ses.LogonId:016x}")
        print(f"    User name:    {ses.UserName}")
        print(f"    Auth Package: {ses.AuthenticationPackage}")
        print(f"    Logon type:   {lot}")
        print(f"    Session:      {ses.Session}")
        print(f"    Logon time:   {fmtt(ses.LogonTime)}")
        print(f"    Logon server: {ses.LogonServer}")
        print(f"    User flags:   0x{ses.UserFlags:x}")

        if not verbose:
            continue

        lf = ses.LastLogonInfo.FailedAttemptCountSinceLastSuccessfulLogon
        print(f"    DNS domain:   {ses.DnsDomainName}")
        print(f"    UPN:          {ses.Upn}")
        print(f"    Last suc lo:  {ses.LastLogonInfo.LastSuccessfulLogon}")
        print(f"    Last fail lo: {ses.LastLogonInfo.LastFailedLogon}")
        print(f"    Failed cnt:   {lf}")
        print(f"    Logon script: {ses.LogonScript}")
        print(f"    Prof. path:   {ses.ProfilePath}")
        print(f"    Home dir:     {ses.HomeDirectory}")
        print(f"    Home drive:   {ses.HomeDirectoryDrive}")
        print(f"    Logoff time:  {fmtt(ses.LogoffTime)}")
        print(f"    Kickoff time: {fmtt(ses.KickOffTime)}")
        print(f"    Pw last set:  {fmtt(ses.PasswordLastSet)}")
        print(f"    Pw can chng:  {fmtt(ses.PasswordCanChange)}")
        print(f"    Pw must chng: {fmtt(ses.PasswordMustChange)}")



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import sys
import shutil
import pathlib
import contextlib
import subprocess

DEFAULT_PREFIX = "$default_prefix$"
prefix = sys.argv[1] if len(sys.argv) >= 2 else DEFAULT_PREFIX
tool = pathlib.Path(__file__).resolve().parent / "rm_sign_tool.py"

def run(*args):
    cmd = [sys.executable, tool]
    cmd.extend(map(str, args))
    subprocess.run(cmd, check=True)

def unlink(pth):
    with contextlib.suppress(OSError):
        pth.unlink()

private = pathlib.Path(f"{prefix}.private.key")
public = pathlib.Path(f"{prefix}.public.key")
signature = pathlib.Path(f"{prefix}.{tool.name}.sig")
fake = pathlib.Path(f"{prefix}.{tool.name}.fake")

try:
    print("going to create keys")
    run("create", private, public)
    print("created keys\n")

    print("going to sign hash")
    run("sign", "-k", private, "-s", signature, tool)
    print("created signature\n")

    print("going to verify hash")
    print("-> ", end="")
    sys.stdout.flush()
    run("verify", "-k", public, "-s", signature, tool)
    print("done verification\n")

    print("demonstrating verification failure")
    print("-> ", end="")
    sys.stdout.flush()
    shutil.copyfile(tool, fake)
    with open(fake, "at") as f:
        f.write("\0")
    run("verify", "-k", public, "-s", signature, fake)
    print("done demonstration\n")
finally:
    if prefix == DEFAULT_PREFIX:
        unlink(private)
        unlink(public)
        unlink(signature)
        unlink(fake)



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import ctypes
from .wtypes import (
    byte_buffer,
    BOOL,
    DWORD,
    PDWORD,
    PPVOID,
    PUINT,
    PVOID,
    PWSTR,
    UINT,
    WORD,
    )
from . import (
    ref,
    raise_on_zero,
    suppress_winerr,
    fun_fact,
    ns_from_struct,
    ERROR_RESOURCE_TYPE_NOT_FOUND,
    )

_ver = ctypes.WinDLL("version.dll", use_last_error=True)

################################################################################

class VS_FIXEDFILEINFO(ctypes.Structure):
    _fields_ = (
        ("dwSignature", DWORD),
        ("dwStrucVersion", DWORD),
        ("dwFileVersionMS", DWORD),
        ("dwFileVersionLS", DWORD),
        ("dwProductVersionMS", DWORD),
        ("dwProductVersionLS", DWORD),
        ("dwFileFlagsMask", DWORD),
        ("dwFileFlags", DWORD),
        ("dwFileOS", DWORD),
        ("dwFileType", DWORD),
        ("dwFileSubtype", DWORD),
        ("dwFileDateMS", DWORD),
        ("dwFileDateLS", DWORD),
        )

################################################################################

_GetFileVersionInfoSize = fun_fact(
    _ver.GetFileVersionInfoSizeW, (DWORD, PWSTR, PDWORD)
    )

def GetFileVersionInfoSize(fname):
    dummy = DWORD()
    res = _GetFileVersionInfoSize(fname, ref(dummy))
    raise_on_zero(res)
    return res

################################################################################

_GetFileVersionInfo = fun_fact(
    _ver.GetFileVersionInfoW, (BOOL, PWSTR, DWORD, DWORD, PVOID)
    )

def GetFileVersionInfo(fname, size=0):
    if size == 0:
        size = GetFileVersionInfoSize(fname)
    buf = byte_buffer(size)
    raise_on_zero(_GetFileVersionInfo(fname, 0, size, buf))
    return buf.raw

################################################################################

_VerQueryValue = fun_fact(
    _ver.VerQueryValueW, (BOOL, PVOID, PWSTR, PPVOID, PUINT)
    )

def VerQueryValue(block, subblock):
    value = PVOID()
    size = UINT()
    raise_on_zero(_VerQueryValue(block, subblock, ref(value), ref(size)))
    if "StringFileInfo" in subblock:
        return ctypes.wstring_at(value.value, size.value).strip("\0")
    return ctypes.string_at(value.value, size.value)

################################################################################

def get_binary_info(fname, block=None):
    if block is None:
        block = GetFileVersionInfo(fname)
    return ns_from_struct(
        VS_FIXEDFILEINFO.from_buffer_copy(VerQueryValue(block, "\\"))
        )

################################################################################

_str_info_names = [
    "CompanyName",
    "ProductName",
    "ProductVersion",
    "FileDescription",
    "FileVersion",
    "InternalName",
    "OriginalFilename",
    "LegalCopyright",
    "LegalTrademarks",
    "Comments",
    "PrivateBuild",
    "SpecialBuild",
    ]

class _LANG_CP(ctypes.Structure):
    _fields_ = (
        ("lang", WORD),
        ("cp", WORD),
        )

################################################################################

def get_string_info(fname, block=None):

    if block is None:
        block = GetFileVersionInfo(fname)

    lc_array = VerQueryValue(block, r"\VarFileInfo\Translation")
    if (len(lc_array) % ctypes.sizeof(_LANG_CP)) != 0:
        raise ValueError("invalid version resource")

    # just take the first language
    lc = _LANG_CP.from_buffer_copy(lc_array)
    lang = f"\\StringFileInfo\\{lc.lang:04x}{lc.cp:04x}\\"

    result = {}
    for k in _str_info_names:
        with suppress_winerr(ERROR_RESOURCE_TYPE_NOT_FOUND):
            result[k] = VerQueryValue(block, lang + k)
    return result

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import ctypes
from .wtypes import (
    BOOL,
    DWORD,
    HANDLE,
    LONG,
    PDWORD,
    POINTER,
    PVOID,
    PWSTR,
    )
from . import (
    ref,
    fun_fact,
    raise_on_zero,
    ns_from_struct,
    )

################################################################################

_wts = ctypes.WinDLL("wtsapi32.dll", use_last_error=True)

class WTS_SESSION_INFO(ctypes.Structure):
    _fields_ = (
        ("SessionId", DWORD),
        ("pWinStationName", PWSTR),
        ("State", LONG),
        )
P_SESSION_INFO = POINTER(WTS_SESSION_INFO)
PP_SESSION_INFO = POINTER(P_SESSION_INFO)

################################################################################

_WTSFreeMemory = fun_fact(_wts.WTSFreeMemory, (None, PVOID))

################################################################################

_WTSEnumerateSessions = fun_fact(
    _wts.WTSEnumerateSessionsW,
    (BOOL, HANDLE, DWORD, DWORD, PP_SESSION_INFO, PDWORD)
    )

def WTSEnumerateSessions(server=None):
    info = P_SESSION_INFO()
    count = DWORD()
    raise_on_zero(_WTSEnumerateSessions(server, 0, 1, ref(info), ref(count)))
    try:
        return tuple(ns_from_struct(info[i]) for i in range(count.value))
    finally:
        _WTSFreeMemory(info)

################################################################################

_WTSDisconnectSession = fun_fact(
    _wts.WTSDisconnectSession, (BOOL, HANDLE, DWORD, BOOL)
    )

def WTSDisconnectSession(session_id, server=None, wait=True):
    raise_on_zero(_WTSDisconnectSession(server, session_id, wait))

################################################################################




################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import tempfile
import pathlib

import ctypes
from .wtypes import (
    BOOL,
    CallbackContext,
    CallbackContextPtr,
    DWORD,
    HINSTANCE,
    HRESULT,
    LONG,
    LPARAM,
    HANDLE,
    HWND,
    INT,
    PBOOL,
    PINT,
    POINTER,
    PWSTR,
    UINT,
    WPARAM,
    )
from . import (
    ref,
    kernel,
    raise_on_zero,
    raise_on_hr,
    fun_fact,
    S_OK,
    TDF_POSITION_RELATIVE_TO_WINDOW,
    )

################################################################################

def _load_comctl():

    # The version of comctl32 that is loaded by LoadLibrary is determined by
    # the activation context that is currently active. Since we need at least
    # version 6, we have to ensure that the current activation context is set
    # up for this version. Currenly the executables 'python.exe' and
    # 'pythonw.exe' contain a manifest resource that ensures that such a context
    # context is active (py 3.10). To be on the safe side, we are going to
    # activate a matching context anyway.

    # create a temporary manifest file
    manifest = (
        "<assembly xmlns='urn:schemas-microsoft-com:asm.v1' manifestVersion=",
        "'1.0'><dependency><dependentAssembly><assemblyIdentity type='win32'",
        " name='Microsoft.Windows.Common-Controls' version='6.0.0.0' ",
        "processorArchitecture='*' publicKeyToken='6595b64144ccf1df' ",
        "language='*'/></dependentAssembly></dependency></assembly>"
        )

    # simply doing the following would lead to a sharing violation :-(
    #
    # with tempfile.NamedTemporaryFile() as f:
    #     tmp_name = f.name
    #     f.write("".join(manifest).encode("ascii"))
    #     actx.lpSource = tmp_name
    #     ctx = kernel.CreateActCtx(actx) # <- sharing violation here
    #
    # so it gets a little more complicated...

    tmp_name = None
    try:
        with tempfile.NamedTemporaryFile(delete=False) as f:
            tmp_name = f.name
            f.write("".join(manifest).encode("ascii"))

        # create activation context from manifest file
        actx = kernel.ACTCTX()
        actx.lpSource = tmp_name
        ctx = kernel.CreateActCtx(actx)
    finally:
        # delete temporary file
        if tmp_name is not None:
            pathlib.Path(tmp_name).unlink()

    # activate context, load libray and and release the context
    kernel.ActivateActCtx(ctx)

    # this calls LoadLibrary
    comctl = ctypes.WinDLL("comctl32.dll", use_last_error=True)

    # Do NOT deactivate the context! Just decrement its ref-count by
    # releasing it.
    kernel.ReleaseActCtx(ctx)

    # verify DLL version
    class DLLVERSIONINFO(ctypes.Structure):
        _fields_ = (
            ("cbSize", DWORD),
            ("dwMajorVersion", DWORD),
            ("dwMinorVersion", DWORD),
            ("dwBuildNumber", DWORD),
            ("dwPlatformID", DWORD),
            )
    dvi = DLLVERSIONINFO()
    dvi.cbSize = ctypes.sizeof(dvi)
    raise_on_hr(comctl.DllGetVersion(ref(dvi)))
    if dvi.dwMajorVersion < 6:
        raise OSError("need at least version 6 of comctl32")

    # register window classes
    class INITCOMMONCONTROLSEX(ctypes.Structure):
        _fields_ = (
            ("dwSize", DWORD),
            ("dwICC", DWORD),
            )
    icc = INITCOMMONCONTROLSEX()
    icc.dwSize = ctypes.sizeof(icc)
    icc.dwICC = 0xffff
    raise_on_zero(comctl.InitCommonControlsEx(ref(icc)))

    return comctl

_ctl = _load_comctl()
del _load_comctl

################################################################################

_TaskDialog = fun_fact(
    _ctl.TaskDialog, (
        HRESULT,
        HWND,
        HINSTANCE,
        PWSTR,
        PWSTR,
        PWSTR,
        DWORD,
        PWSTR,
        PINT
        )
    )

def TaskDialog(owner, title, main_instr, content, buttons, icon, inst=None):
    res = INT()
    raise_on_hr(
        _TaskDialog(
            owner,
            inst,
            title,
            main_instr,
            content,
            buttons,
            icon,
            ref(res)
            )
        )
    return res.value

################################################################################

_TaskDialogCallback = ctypes.WINFUNCTYPE(
    LONG,
    HWND,
    UINT,
    WPARAM,
    LPARAM,
    CallbackContextPtr
    )

@_TaskDialogCallback
def _TskDlgCb(hwnd, msg, wp, lp, ctxt):
    # cannot propagate exceptions from callback
    with kernel.terminate_on_exception():
        cbc = ctxt.contents
        res = cbc.callback(hwnd, msg, wp, lp, cbc.context)
        # return S_OK if the callback fails to return a value
        return S_OK if res is None else res

################################################################################

# ALL TASKDIALOG STRUCTURES NEED AN ALIGNMENT OF 1 (_pack_ = 1)!!!

class TASKDIALOG_BUTTON(ctypes.Structure):
    _pack_ = 1
    _fields_ = (
        ("nButtonID", INT),
        ("pszButtonText", PWSTR),
        )

PTASKDIALOG_BUTTON = POINTER(TASKDIALOG_BUTTON)

class _TD_MAIN_ICON(ctypes.Union):
    _pack_ = 1
    _fields_ = (("hMainIcon", HANDLE), ("pszMainIcon", PWSTR))

class _TD_FOOTER_ICON(ctypes.Union):
    _pack_ = 1
    _fields_ = (("hFooterIcon", HANDLE), ("pszFooterIcon", PWSTR))

class TASKDIALOGCONFIG(ctypes.Structure):
    _pack_ = 1
    _anonymous_ = ("_main_icon", "_footer_icon")
    _fields_ = (
        ("cbSize", UINT),
        ("hwndParent", HWND),
        ("hInstance", HINSTANCE),
        ("dwFlags", DWORD),
        ("dwCommonButtons", DWORD),
        ("pszWindowTitle", PWSTR),
        ("_main_icon", _TD_MAIN_ICON),
        ("pszMainInstruction", PWSTR),
        ("pszContent", PWSTR),
        ("cButtons", UINT),
        ("pButtons", PTASKDIALOG_BUTTON),
        ("nDefaultButton", INT),
        ("cRadioButtons", UINT),
        ("pRadioButtons", PTASKDIALOG_BUTTON),
        ("nDefaultRadioButton", INT),
        ("pszVerificationText", PWSTR),
        ("pszExpandedInformation", PWSTR),
        ("pszExpandedControlText", PWSTR),
        ("pszCollapsedControlText", PWSTR),
        ("_footer_icon", _TD_FOOTER_ICON),
        ("pszFooter", PWSTR),
        ("pfCallback", _TaskDialogCallback),
        ("lpCallbackData", CallbackContextPtr),
        ("cxWidth", UINT),
        )

    def __init__(self):
        self.cbSize = ctypes.sizeof(self)

PTASKDIALOGCONFIG = POINTER(TASKDIALOGCONFIG)

################################################################################

_TaskDialogIndirect = fun_fact(
    _ctl.TaskDialogIndirect, (
        HRESULT,
        PTASKDIALOGCONFIG,
        PINT,
        PINT,
        PBOOL
        )
    )

def TaskDialogIndirect(tsk_dlg_cfg):
    button_idx = INT()
    radio_idx = INT()
    verified = BOOL()
    raise_on_hr(
        _TaskDialogIndirect(
            ref(tsk_dlg_cfg),
            ref(button_idx),
            ref(radio_idx),
            ref(verified)
            )
        )
    return button_idx.value, radio_idx.value, bool(verified.value)

################################################################################

def tsk_dlg_callback(tsk_dlg_cfg, callback, context=None):
    ctxt = CallbackContext(callback, context)
    tsk_dlg_cfg.pfCallback = _TskDlgCb
    tsk_dlg_cfg.lpCallbackData = ctypes.pointer(ctxt)
    return TaskDialogIndirect(tsk_dlg_cfg)

################################################################################

def tsk_dlg_centered(owner, title, instr, content, buttons, icon, inst=None):
    tdc = TASKDIALOGCONFIG()
    tdc.hwndParent = owner
    tdc.pszWindowTitle = title
    tdc.pszMainInstruction = instr
    tdc.pszContent = content
    tdc.dwCommonButtons = buttons
    tdc.pszMainIcon = icon
    tdc.hInstance = inst
    tdc.dwFlags = TDF_POSITION_RELATIVE_TO_WINDOW
    button_idx = INT()
    raise_on_hr(_TaskDialogIndirect(ref(tdc), ref(button_idx), None, None))
    return button_idx.value

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import re
import ctypes
from .wtypes import (
    byte_buffer,
    string_buffer,
    BOOL,
    DWORD,
    GUID,
    HANDLE,
    HWND,
    PBYTE,
    PDWORD,
    PGUID,
    POINTER,
    PTR_HAS_64_BITS,
    PVOID,
    PWSTR,
    ScdToBeClosed,
    ULONG_PTR,
    WCHAR,
    )
from . import (
    ref,
    raise_on_zero,
    fun_fact,
    suppress_winerr,
    INVALID_HANDLE_VALUE,
    DIGCF_PRESENT,
    DIGCF_ALLCLASSES,
    DIGCF_DEVICEINTERFACE,
    DICS_ENABLE,
    DICS_DISABLE,
    DICS_FLAG_CONFIGSPECIFIC,
    DIF_PROPERTYCHANGE,
    SPDRP_DEVICEDESC,
    ERROR_NO_MORE_ITEMS,
    ERROR_NOT_FOUND,
    )
from .cfgmgr import (
    CM_Get_Device_ID,
    CM_Get_DevNode_Status,
    CM_Enumerate_Enumerators,
    CM_Enumerate_Classes,
    )
from .advapi import registry_to_py

_sua = ctypes.WinDLL("setupapi.dll", use_last_error=True)

# SetupAPI.h packs structures differently in 64 and 32 bit mode!
_SUA_PACK = 8 if PTR_HAS_64_BITS else 1

################################################################################

class SP_DEVINFO_DATA(ctypes.Structure):
    _pack_ = _SUA_PACK
    _fields_ = (
        ("cbSize", DWORD),
        ("ClassGuid", GUID),
        ("DevInst", DWORD),
        ("Reserved", ULONG_PTR),
        )

    def __init__(self):
        self.cbSize = ctypes.sizeof(self)

PSP_DEVINFO_DATA = POINTER(SP_DEVINFO_DATA)

################################################################################

class SP_CLASSINSTALL_HEADER(ctypes.Structure):
    _pack_ = _SUA_PACK
    _fields_ = (
        ("cbSize", DWORD),
        ("InstallFunction", DWORD),
        )

################################################################################

class SP_PROPCHANGE_PARAMS(ctypes.Structure):
    _pack_ = _SUA_PACK
    _fields_ = (
        ("ClassInstallHeader", SP_CLASSINSTALL_HEADER),
        ("StateChange", DWORD),
        ("Scope", DWORD),
        ("HwProfile", DWORD),
        )

    def __init__(self, func, change, scope=DICS_FLAG_CONFIGSPECIFIC, prof=0):
        self.ClassInstallHeader.cbSize = ctypes.sizeof(self.ClassInstallHeader)
        self.ClassInstallHeader.InstallFunction = func
        self.StateChange = change
        self.Scope = scope
        self.HwProfile = prof

PSP_PROPCHANGE_PARAMS = POINTER(SP_PROPCHANGE_PARAMS)

################################################################################

class SP_DEVICE_INTERFACE_DATA(ctypes.Structure):
    _pack_ = _SUA_PACK
    _fields_ = (
        ("cbSize", DWORD),
        ("InterfaceClassGuid", GUID),
        ("Flags", DWORD),
        ("Reserved", ULONG_PTR),
        )

    def __init__(self):
        self.cbSize = ctypes.sizeof(self)

PSP_DEVICE_INTERFACE_DATA = POINTER(SP_DEVICE_INTERFACE_DATA)

################################################################################

class SP_DEVICE_INTERFACE_DETAIL_DATA(ctypes.Structure):
    _pack_ = _SUA_PACK
    _fields_ = (
        ("cbSize", DWORD),
        ("DevicePath", WCHAR * 1),
       )

def make_dev_iface_detail(size):
    buf = byte_buffer(size)
    ifd = SP_DEVICE_INTERFACE_DETAIL_DATA.from_buffer(buf)
    ifd.cbSize = ctypes.sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA)
    offs = SP_DEVICE_INTERFACE_DETAIL_DATA.DevicePath.offset
    return buf, ctypes.addressof(buf) + offs

################################################################################

_SetupDiDestroyDeviceInfoList = fun_fact(
    _sua.SetupDiDestroyDeviceInfoList, (BOOL, HANDLE)
    )

def SetupDiDestroyDeviceInfoList(info_set):
    raise_on_zero(_SetupDiDestroyDeviceInfoList(info_set))

################################################################################

class HDEVINFO(
        ScdToBeClosed,
        HANDLE,
        close_func=SetupDiDestroyDeviceInfoList,
        invalid=INVALID_HANDLE_VALUE
        ):
    pass

################################################################################

_SetupDiGetClassDevs = fun_fact(
    _sua.SetupDiGetClassDevsW,
    (HANDLE, PGUID, PWSTR, HWND, DWORD)
    )

def SetupDiGetClassDevs(
        guid=None,
        enumerator=None,
        flags=DIGCF_PRESENT | DIGCF_ALLCLASSES,
        hwnd=None
        ):
    if guid is not None:
        guid = ref(GUID(guid))  # need ctypes instance
        flags &= ~ DIGCF_ALLCLASSES
    res = HDEVINFO(_SetupDiGetClassDevs(guid, enumerator, hwnd, flags))
    res.raise_on_invalid()
    return res

################################################################################

_SetupDiEnumDeviceInfo = fun_fact(
    _sua.SetupDiEnumDeviceInfo,
    (BOOL, HANDLE, DWORD, PSP_DEVINFO_DATA)
    )

def SetupDiEnumDeviceInfo(info_set, idx, deinda):
    return _SetupDiEnumDeviceInfo(info_set, idx, deinda)

################################################################################

def get_device_enumerators():
    res = []
    idx = 0
    with suppress_winerr(ERROR_NOT_FOUND):
        while True:
            res.append(CM_Enumerate_Enumerators(idx))
            idx += 1
    return res

################################################################################

def get_device_classes(flags=0):
    res = []
    idx = 0
    with suppress_winerr(ERROR_NOT_FOUND):
        while True:
            res.append(CM_Enumerate_Classes(idx, flags))
            idx += 1
    return res

################################################################################

_SetupDiClassNameFromGuid = fun_fact(
    _sua.SetupDiClassNameFromGuidW,
    (BOOL, PGUID, PWSTR, DWORD, PDWORD)
    )

def SetupDiClassNameFromGuid(guid):
    guid = GUID(guid)  # need ctypes instance
    req_size = DWORD(0)
    _SetupDiClassNameFromGuid(ref(guid), None, 0, ref(req_size))
    name = string_buffer(req_size.value)
    raise_on_zero(
        _SetupDiClassNameFromGuid(
            ref(guid),
            name,
            req_size.value,
            ref(req_size)
            )
        )
    return name.value

################################################################################

_SetupDiRemoveDevice = fun_fact(
    _sua.SetupDiRemoveDevice, (BOOL, HANDLE, PSP_DEVINFO_DATA)
    )

def SetupDiRemoveDevice(info_set, deinda):
    raise_on_zero(_SetupDiRemoveDevice(info_set, deinda))

################################################################################

_SetupDiCreateDeviceInfoList = fun_fact(
    _sua.SetupDiCreateDeviceInfoList, (HANDLE, PGUID, HWND)
    )

def SetupDiCreateDeviceInfoList(guid=None, hwnd=None):
    if guid is not None:
        guid = ref(GUID(guid))  # need ctypes instance
    res = HDEVINFO(_SetupDiCreateDeviceInfoList(guid, hwnd))
    res.raise_on_invalid()
    return res

################################################################################

_SetupDiGetDeviceInstanceId = fun_fact(
    _sua.SetupDiGetDeviceInstanceIdW,
    (BOOL, HANDLE, PSP_DEVINFO_DATA, PWSTR, DWORD, PDWORD)
    )

def SetupDiGetDeviceInstanceId(info_set, deinda):
    req_size = DWORD()
    _SetupDiGetDeviceInstanceId(info_set, deinda, None, 0, ref(req_size))
    idstr = string_buffer(req_size.value)
    raise_on_zero(
        _SetupDiGetDeviceInstanceId(
            info_set,
            deinda,
            idstr,
            req_size.value,
            ref(req_size)
            )
        )
    return idstr.value

################################################################################

_SetupDiOpenDeviceInfo = fun_fact(
    _sua.SetupDiOpenDeviceInfoW,
    (BOOL, HANDLE, PWSTR, HWND, DWORD, PSP_DEVINFO_DATA)
    )

def SetupDiOpenDeviceInfo(info_set, inst_id, hwnd=None, flags=0, p_info=None):
    raise_on_zero(
        _SetupDiOpenDeviceInfo(info_set, inst_id, hwnd, flags, p_info)
        )

################################################################################

_SetupDiClassGuidsFromNameEx = fun_fact(
    _sua.SetupDiClassGuidsFromNameExW, (
        BOOL,
        PWSTR,
        PGUID,
        DWORD,
        PDWORD,
        PWSTR,
        PVOID
        )
    )

def SetupDiClassGuidsFromNameEx(cls_name, machine_name=None):
    req_size = DWORD()
    _SetupDiClassGuidsFromNameEx(
        cls_name,
        None,
        0,
        ref(req_size),
        machine_name,
        None
        )
    CLASS_GUIDS = GUID * req_size.value
    guids = CLASS_GUIDS()
    raise_on_zero(
        _SetupDiClassGuidsFromNameEx(
            cls_name,
            ctypes.cast(guids, PGUID),
            req_size.value,
            ref(req_size),
            machine_name,
            None
            )
        )
    return [g.uuid() for g in guids]

################################################################################

def enum_dev_ids(guid=None, enumerator=None, flags=None, rx=None):
    if flags is None:
        flags = DIGCF_PRESENT | DIGCF_ALLCLASSES
    if isinstance(rx, str):
        rx = re.compile(rx)

    with SetupDiGetClassDevs(guid, enumerator, flags) as devs:
        deinda = SP_DEVINFO_DATA()
        idx = 0
        res = []
        while SetupDiEnumDeviceInfo(devs, idx, ref(deinda)):
            did = CM_Get_Device_ID(deinda.DevInst)
            if rx is None or rx.search(did):
                res.append(did)
            idx += 1
    return res

################################################################################

def build_info_set(guid=None, enumerator=None, flags=None, rx=None):
    selected = SetupDiCreateDeviceInfoList()
    if guid is None and enumerator is None and rx is None:
        return selected

    if flags is None:
        flags = DIGCF_PRESENT | DIGCF_ALLCLASSES
    if isinstance(rx, str):
        rx = re.compile(rx)

    with SetupDiGetClassDevs(guid, enumerator, flags) as enum_devs:
        deinda = SP_DEVINFO_DATA()
        idx = 0
        while SetupDiEnumDeviceInfo(enum_devs, idx, ref(deinda)):
            did = CM_Get_Device_ID(deinda.DevInst)
            if rx is None or rx.search(did):
                SetupDiOpenDeviceInfo(selected, did)
            idx += 1
    return selected

################################################################################

def get_non_present_info_set():
    non_present = SetupDiCreateDeviceInfoList()
    with SetupDiGetClassDevs(flags=DIGCF_ALLCLASSES) as all_devs:
        deinda = SP_DEVINFO_DATA()
        idx = 0
        while SetupDiEnumDeviceInfo(all_devs, idx, ref(deinda)):
            # If the device isn't currently present (as indicated by
            # failing to retrieve its status), then add it to the list.
            try:
                CM_Get_DevNode_Status(deinda.DevInst)
            except OSError:
                SetupDiOpenDeviceInfo(
                    non_present,
                    CM_Get_Device_ID(deinda.DevInst)
                    )
            idx += 1
    return non_present

################################################################################

def enum_info_set(info_set, cleanup=True):
    deinda = SP_DEVINFO_DATA()
    idx = 0
    try:
        while SetupDiEnumDeviceInfo(info_set, idx, ref(deinda)):
            yield info_set, deinda
            idx += 1
    finally:
        if cleanup:
            SetupDiDestroyDeviceInfoList(info_set)

################################################################################

def remove_non_present_devices():
    for iset, ddat in enum_info_set(get_non_present_info_set()):
        SetupDiRemoveDevice(iset, ddat)

################################################################################

_SetupDiSetClassInstallParams = fun_fact(
    _sua.SetupDiSetClassInstallParamsW,
    (BOOL, HANDLE, PSP_DEVINFO_DATA, PSP_PROPCHANGE_PARAMS, DWORD)
    )

def SetupDiSetClassInstallParams(info_set, deinda, pparams):
    raise_on_zero(
        _SetupDiSetClassInstallParams(
            info_set,
            deinda,
            pparams,
            ctypes.sizeof(SP_PROPCHANGE_PARAMS)
            )
        )

################################################################################

_SetupDiCallClassInstaller = fun_fact(
    _sua.SetupDiCallClassInstaller,
    (BOOL, DWORD, HANDLE, PSP_DEVINFO_DATA)
    )

def SetupDiCallClassInstaller(func, info_set, deinda):
    raise_on_zero(_SetupDiCallClassInstaller(func, info_set, deinda))

################################################################################

def _prop_change(
        info_set,
        deinda,
        change,
        scope=DICS_FLAG_CONFIGSPECIFIC,
        prof=0
        ):
    params = SP_PROPCHANGE_PARAMS(DIF_PROPERTYCHANGE, change, scope, prof)
    SetupDiSetClassInstallParams(info_set, deinda, ref(params))
    SetupDiCallClassInstaller(DIF_PROPERTYCHANGE, info_set, deinda)

################################################################################

def _change_devices(change, guid=None, enumerator=None, rx=None):
    with build_info_set(guid, enumerator, None, rx) as info_set:
        deinda = SP_DEVINFO_DATA()
        idx = 0
        while SetupDiEnumDeviceInfo(info_set, idx, ref(deinda)):
            _prop_change(info_set, deinda, change)
            idx += 1

################################################################################

def enable_devices(guid=None, enumerator=None, rx=None):
    _change_devices(DICS_ENABLE, guid, enumerator, rx)

################################################################################

def disable_devices(guid=None, enumerator=None, rx=None):
    _change_devices(DICS_DISABLE, guid, enumerator, rx)

################################################################################

_SetupDiGetDeviceRegistryProperty = fun_fact(
    _sua.SetupDiGetDeviceRegistryPropertyW, (
        BOOL,
        HANDLE,
        PSP_DEVINFO_DATA,
        DWORD,
        PDWORD,
        PBYTE,
        DWORD,
        PDWORD
        )
    )

def SetupDiGetDeviceRegistryProperty(info_set, deinda, prop):
    reg_type = DWORD()
    req_size = DWORD()
    _SetupDiGetDeviceRegistryProperty(
        info_set,
        deinda,
        prop,
        ref(reg_type),
        None,
        0,
        ref(req_size)
        )
    buf = byte_buffer(req_size.value)
    raise_on_zero(
        _SetupDiGetDeviceRegistryProperty(
            info_set,
            deinda,
            prop,
            ref(reg_type),
            ctypes.cast(buf, PBYTE),
            req_size,
            ref(req_size)
            )
        )
    return registry_to_py(reg_type.value, buf.raw[:req_size.value])

################################################################################

def desc_from_info_set(info_set, deinda):
    return SetupDiGetDeviceRegistryProperty(
        info_set,
        deinda,
        SPDRP_DEVICEDESC
        )[0]

################################################################################

_SetupDiEnumDeviceInterfaces = fun_fact(
    _sua.SetupDiEnumDeviceInterfaces, (
        BOOL,
        HANDLE,
        PSP_DEVINFO_DATA,
        PGUID,
        DWORD,
        PSP_DEVICE_INTERFACE_DATA
        )
    )

def SetupDiEnumDeviceInterfaces(info_set, guid, idx):
    did = SP_DEVICE_INTERFACE_DATA()
    raise_on_zero(
        _SetupDiEnumDeviceInterfaces(
            info_set,
            None,
            ref(GUID(guid)),  # need ctypes instance
            idx,
            ref(did)
            )
        )
    return did

################################################################################

def enum_dev_interfaces(guid):
    flags = DIGCF_PRESENT | DIGCF_DEVICEINTERFACE
    with SetupDiGetClassDevs(flags=flags, guid=guid) as info_set:
        idx = 0
        with suppress_winerr(ERROR_NO_MORE_ITEMS):
            while True:
                did = SetupDiEnumDeviceInterfaces(info_set, guid, idx)
                idx += 1
                yield info_set, did

################################################################################

_SetupDiGetDeviceInterfaceDetail = fun_fact(
    _sua.SetupDiGetDeviceInterfaceDetailW, (
        BOOL,
        HANDLE,
        PSP_DEVICE_INTERFACE_DATA,
        PVOID,
        DWORD,
        PDWORD,
        PSP_DEVINFO_DATA
        )
    )

def SetupDiGetDeviceInterfaceDetail(info_set, did):
    req_size = DWORD(0)
    _SetupDiGetDeviceInterfaceDetail(
        info_set,
        ref(did),
        None,
        0,
        ref(req_size),
        None
        )

    buf, addr = make_dev_iface_detail(req_size.value)
    deinda = SP_DEVINFO_DATA()
    raise_on_zero(
        _SetupDiGetDeviceInterfaceDetail(
            info_set,
            ref(did),
            ref(buf),
            req_size,
            ref(req_size),
            ref(deinda),
            )
        )
    return ctypes.wstring_at(addr), deinda

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

from types import SimpleNamespace as _namespace
from datetime import datetime as _dt
from datetime import UTC as _UTC
import ctypes

from .wtypes import (
    byte_buffer,
    string_buffer,
    ArgcArgvFromArgs,
    BOOL,
    BYTE,
    DWORD,
    ENDIANNESS,
    FILETIME,
    HANDLE,
    INT,
    LARGE_INTEGER,
    LONG,
    LUID,
    PBOOL,
    PBYTE,
    PDWORD,
    PFILETIME,
    PHANDLE,
    PLUID,
    POINTER,
    PPWSTR,
    PPVOID,
    PVOID,
    PWSTR,
    ScdToBeClosed,
    WCHAR_SIZE,
    WinError,
    WORD,
    )
from . import (
    cmdline_from_args,
    CRED_TYPE_GENERIC,
    DACL_SECURITY_INFORMATION,
    ERROR_HANDLE_EOF,
    ERROR_INSUFFICIENT_BUFFER,
    ERROR_MORE_DATA,
    ERROR_NOT_ALL_ASSIGNED,
    ERROR_NOT_FOUND,
    ERROR_NO_MORE_ITEMS,
    EVENTLOG_BACKWARDS_READ,
    EVENTLOG_SEQUENTIAL_READ,
    fun_fact,
    GROUP_SECURITY_INFORMATION,
    KEY_ALL_ACCESS,
    KEY_READ,
    KEY_WOW64_64KEY,
    ns_from_struct,
    MAXIMUM_ALLOWED,
    OWNER_SECURITY_INFORMATION,
    PROCESS_QUERY_LIMITED_INFORMATION,
    raise_if,
    raise_on_err,
    raise_on_zero,
    ref,
    REG_DWORD,
    REG_EXPAND_SZ,
    REG_MULTI_SZ,
    REG_QWORD,
    REG_SZ,
    SACL_SECURITY_INFORMATION,
    SC_ENUM_PROCESS_INFO,
    SC_STATUS_PROCESS_INFO,
    SE_PRIVILEGE_ENABLED,
    suppress_winerr,
    TOKEN_QUERY,
    TokenElevationType,
    TokenElevationTypeFull,
    TokenGroups,
    TokenPrivileges,
    TokenSessionId,
    TokenUser,
    WinBuiltinAdministratorsSid,
    WinLocalSystemSid,
    )
from .kernel import (
    CloseHandle,
    GetCurrentProcess,
    GetLastError,
    get_local_tzinfo,
    KHANDLE,
    LocalFree,
    OpenProcess,
    PSECURITY_ATTRIBUTES,
    PSTARTUPINFO,
    PPROCESS_INFORMATION,
    PROCESS_INFORMATION,
    STARTUPINFO,
    )
from .ntdll import enum_processes

_adv = ctypes.WinDLL("advapi32.dll", use_last_error=True)

################################################################################

# values of predefined keys

_PREDEFINED_KEYS = {
    0x80000000: "HKEY_CLASSES_ROOT",
    0x80000001: "HKEY_CURRENT_USER",
    0x80000002: "HKEY_LOCAL_MACHINE",
    0x80000003: "HKEY_USERS",
    0x80000004: "HKEY_PERFORMANCE_DATA",
    0x80000050: "HKEY_PERFORMANCE_TEXT",
    0x80000060: "HKEY_PERFORMANCE_NLSTEXT",
    0x80000005: "HKEY_CURRENT_CONFIG",
    0x80000006: "HKEY_DYN_DATA",
    0x80000007: "HKEY_CURRENT_USER_LOCAL_SETTINGS",
    }

################################################################################

_RegCloseKey = fun_fact(_adv.RegCloseKey, (LONG, HANDLE))

def RegCloseKey(key):
    raise_on_err(_RegCloseKey(key))

################################################################################

class HKEY(ScdToBeClosed, HANDLE, close_func=RegCloseKey, invalid=0):

    def close(self):
        # predefined keys cannot be closed (ERROR_INVALID_HANDLE)
        if self.value not in _PREDEFINED_KEYS:
            super().close()

PHKEY = POINTER(HKEY)

################################################################################

# predefined keys as instances of HKEY
globals().update((n, HKEY(v)) for v, n in _PREDEFINED_KEYS.items())

# linters won't recognize 'globals().update()' -> noqa
HKCR = HKEY_CLASSES_ROOT    # noqa: F821
HKCU = HKEY_CURRENT_USER    # noqa: F821
HKLM = HKEY_LOCAL_MACHINE   # noqa: F821

################################################################################

def is_registry_string(reg_type):
    return reg_type in (REG_SZ, REG_EXPAND_SZ, REG_MULTI_SZ)

################################################################################

def registry_to_py(reg_type, data):
    if is_registry_string(reg_type):
        if len(data) <= 1:
            result = [] if reg_type == REG_MULTI_SZ else ""
        else:
            if (len(data) & 1) != 0 and data[-1] == 0:
                data = data[:-1]
            result = data.decode("utf-16").strip("\0")
            if reg_type == REG_MULTI_SZ:
                result = result.split("\0")
    elif reg_type in (REG_DWORD, REG_QWORD):
        result = int.from_bytes(data, byteorder=ENDIANNESS, signed=False)
    else:
        result = data

    return result, reg_type

################################################################################

_RegOpenKeyEx = fun_fact(
    _adv.RegOpenKeyExW,
    (LONG, HKEY, PWSTR, DWORD, DWORD, PHKEY)
    )

def RegOpenKeyEx(parent, name, access=KEY_READ):
    key = HKEY()
    raise_on_err(_RegOpenKeyEx(parent, name, 0, access, ref(key)))
    return key

################################################################################

_RegQueryInfoKey = fun_fact(
    _adv.RegQueryInfoKeyW, (
        LONG,
        HKEY,
        PWSTR,
        PDWORD,
        PDWORD,
        PDWORD,
        PDWORD,
        PDWORD,
        PDWORD,
        PDWORD,
        PDWORD,
        PDWORD,
        PFILETIME
        )
    )

def RegQueryInfoKey(key):
    num_sub_keys = DWORD()
    max_sub_key_len = DWORD()
    num_values = DWORD()
    max_value_name_len = DWORD()
    max_value_len = DWORD()
    last_written = FILETIME()
    raise_on_err(
        _RegQueryInfoKey(
            key,
            None,
            None,
            None,
            ref(num_sub_keys),
            ref(max_sub_key_len),
            None,
            ref(num_values),
            ref(max_value_name_len),
            ref(max_value_len),
            None,
            ref(last_written)
            )
        )
    return _namespace(
        num_sub_keys=num_sub_keys.value,
        max_sub_key_len=max_sub_key_len.value,
        num_values=num_values.value,
        max_value_name_len=max_value_name_len.value,
        max_value_len=max_value_len.value,
        last_written=last_written,
        )

################################################################################

_RegCreateKeyEx = fun_fact(
    _adv.RegCreateKeyExW, (
        LONG,
        HKEY,
        PWSTR,
        DWORD,
        PWSTR,
        DWORD,
        DWORD,
        PVOID,
        PHKEY,
        PDWORD,
        )
    )

def RegCreateKeyEx(parent, name, access=KEY_ALL_ACCESS):
    key = HKEY()
    raise_on_err(
        _RegCreateKeyEx(
            parent,
            name,
            0,
            0,
            access,
            None,
            ref(key),
            None
            )
        )
    return key

################################################################################

_RegDeleteKeyEx = fun_fact(
    _adv.RegDeleteKeyExW, (LONG, HKEY, PWSTR, DWORD, DWORD)
    )

def RegDeleteKeyEx(parent, name, access=KEY_WOW64_64KEY):
    raise_on_err(
        _RegDeleteKeyEx(
            parent,
            name,
            access,
            0
            )
        )

################################################################################

_RegDeleteValue = fun_fact(
    _adv.RegDeleteValueW, (LONG, HKEY, PWSTR)
    )

def RegDeleteValue(key, name):
    raise_on_err(_RegDeleteValue(key, name))

################################################################################

_RegDeleteKeyValue = fun_fact(
    _adv.RegDeleteKeyValueW, (LONG, HKEY, PWSTR, PWSTR)
    )

def RegDeleteKeyValue(parent, key_name, value_name):
    raise_on_err(
        _RegDeleteKeyValue(
            parent,
            key_name,
            value_name
            )
        )

################################################################################

# The Windows docs claim that the max key name length is 255 characters, plus
# a terminating null character.  However, empirical testing demonstrates that
# it is possible to create a 256 character key that is missing the terminating
# null.  RegEnumKeyEx requires a 257 character buffer to retrieve such a key
# name.
_MAX_KEY_LEN = 257

################################################################################

_RegEnumKeyEx = fun_fact(
    _adv.RegEnumKeyExW, (
        LONG,
        HKEY,
        DWORD,
        PWSTR,
        PDWORD,
        PDWORD,
        PWSTR,
        PDWORD,
        PFILETIME
        )
    )

def RegEnumKeyEx(key, index):
    name_len = DWORD(_MAX_KEY_LEN)
    name = string_buffer(_MAX_KEY_LEN)
    raise_on_err(
        _RegEnumKeyEx(
            key,
            index,
            name,
            ref(name_len),
            None,
            None,
            None,
            None
            )
        )
    return name.value

################################################################################

def reg_enum_keys(key):
    index = 0
    with suppress_winerr(ERROR_NO_MORE_ITEMS):
        while True:
            sub_key_name = RegEnumKeyEx(key, index)
            index += 1
            yield sub_key_name

################################################################################

_RegEnumValue = fun_fact(
    _adv.RegEnumValueW, (
        LONG,
        HKEY,
        DWORD,
        PWSTR,
        PDWORD,
        PDWORD,
        PDWORD,
        PBYTE,
        PDWORD,
        )
    )

def RegEnumValue(key, index):
    info = RegQueryInfoKey(key)
    nlen = DWORD(info.max_value_name_len + 1)
    vlen = DWORD(info.max_value_len + 1)
    name = string_buffer(nlen.value)
    value = byte_buffer(vlen.value)
    typ = DWORD()
    while True:
        err = _RegEnumValue(
            key,
            index,
            name,
            ref(nlen),
            None,
            ref(typ),
            ctypes.cast(value, PBYTE),
            ref(vlen)
            )
        if err == 0:
            break
        elif err == ERROR_MORE_DATA:
            vlen = DWORD(vlen.value * 2)
            value = byte_buffer(vlen.value)
        else:
            raise WinError(err)

    return (name.value, *registry_to_py(typ.value, value.raw[:vlen.value]))

################################################################################

def reg_enum_values(key):
    index = 0
    with suppress_winerr(ERROR_NO_MORE_ITEMS):
        while True:
            tpl = RegEnumValue(key, index)
            index += 1
            yield tpl

################################################################################

_RegQueryValueEx = fun_fact(
    _adv.RegQueryValueExW, (
        LONG,
        HKEY,
        PWSTR,
        PDWORD,
        PDWORD,
        PBYTE,
        PDWORD
        )
    )

def RegQueryValueEx(key, name):
    vlen = DWORD(256)
    value = byte_buffer(vlen.value)
    typ = DWORD()
    while True:
        err = _RegQueryValueEx(
            key,
            name,
            None,
            ref(typ),
            ctypes.cast(value, PBYTE),
            ref(vlen)
            )
        if err == 0:
            break
        elif err == ERROR_MORE_DATA:
            vlen = DWORD(vlen.value * 2)
            value = byte_buffer(vlen.value)
        else:
            raise WinError(err)

    return registry_to_py(typ.value, value[:vlen.value])

################################################################################

_RegSetValueEx = fun_fact(
    _adv.RegSetValueExW, (
        LONG,
        HKEY,
        PWSTR,
        DWORD,
        DWORD,
        PVOID,
        DWORD,
        )
    )

def RegSetValueEx(key, name, typ, data):
    dta = byte_buffer(data)
    raise_on_err(
        _RegSetValueEx(
            key,
            name,
            0,
            typ,
            ref(dta),
            len(data)
            )
        )

################################################################################

_RegSetKeyValue = fun_fact(
    _adv.RegSetKeyValueW, (
        LONG,
        HKEY,
        PWSTR,
        PWSTR,
        DWORD,
        PVOID,
        DWORD,
        )
    )

def RegSetKeyValue(parent, key_name, value_name, typ, data):
    dta = byte_buffer(data)
    raise_on_err(
        _RegSetKeyValue(
            parent,
            key_name,
            value_name,
            typ,
            ref(dta),
            len(data)
            )
        )

################################################################################

def reg_set_str(key, name, string, typ=None):
    typ = REG_SZ if typ is None else typ
    if not is_registry_string(typ):
        raise ValueError(f"invalid registry type: {typ}")
    value = string_buffer(string, len(string))
    raise_on_err(
        _RegSetValueEx(
            key,
            name,
            0,
            typ,
            ref(value),
            ctypes.sizeof(value)
            )
        )

################################################################################

def reg_set_dword(key, name, dword):
    size = ctypes.sizeof(DWORD)
    data = dword.to_bytes(size, ENDIANNESS)
    raise_on_err(
        _RegSetValueEx(
            key,
            name,
            0,
            REG_DWORD,
            ref(data),
            size
            )
        )

################################################################################

_RegFlushKey = fun_fact(_adv.RegFlushKey, (LONG, HKEY))

def RegFlushKey(key):
    raise_on_err(_RegFlushKey(key))

################################################################################

_RegLoadAppKey = fun_fact(
    _adv.RegLoadAppKeyW,
    (LONG, PWSTR, PHKEY, DWORD, DWORD, DWORD)
    )

REG_PROCESS_APPKEY = 1

def RegLoadAppKey(hive_name, acc=KEY_ALL_ACCESS, opt=REG_PROCESS_APPKEY):
    key = HKEY()
    raise_on_err(_RegLoadAppKey(hive_name, ref(key), acc, opt, 0))
    key.raise_on_invalid()
    return key

################################################################################

_IsValidSid = fun_fact(_adv.IsValidSid, (BOOL, PVOID))

def IsValidSid(psid):
    return _IsValidSid(psid) != 0

################################################################################

_GetLengthSid = fun_fact(_adv.GetLengthSid, (DWORD, PVOID))

def GetLengthSid(psid):
    if not IsValidSid(psid):
        raise ValueError(f"invalid SID: {psid:x}")
    return _GetLengthSid(psid)

################################################################################

_ConvertStringSidToSid = fun_fact(
    _adv.ConvertStringSidToSidW, (BOOL, PWSTR, PVOID)
    )

def ConvertStringSidToSid(string_sid):
    sid = PVOID()
    try:
        raise_on_zero(_ConvertStringSidToSid(string_sid, ref(sid)))
        return ctypes.string_at(sid, GetLengthSid(sid))
    finally:
        LocalFree(sid)

################################################################################

_ConvertSidToStringSid = fun_fact(
    _adv.ConvertSidToStringSidW, (BOOL, PVOID, PPWSTR)
    )

def ConvertSidToStringSid(sid):
    bin_sid = byte_buffer(sid)
    str_sid = PWSTR()
    try:
        raise_on_zero(_ConvertSidToStringSid(ref(bin_sid), ref(str_sid)))
        return ctypes.wstring_at(str_sid)
    finally:
        LocalFree(str_sid)

################################################################################

_CreateWellKnownSid = fun_fact(
    _adv.CreateWellKnownSid, (BOOL, INT, PVOID, PVOID, PDWORD)
    )

def CreateWellKnownSid(sid_type, domain=None):
    if domain is not None:
       domain = byte_buffer(domain)
    size = DWORD(0)
    _CreateWellKnownSid(sid_type, domain, None, ref(size))
    wks = byte_buffer(size.value)
    raise_on_zero(_CreateWellKnownSid(sid_type, domain, wks, ref(size)))
    return wks.raw[:size.value]

################################################################################

_IsWellKnownSid = fun_fact(
    _adv.IsWellKnownSid, (BOOL, PVOID, INT)
    )

def IsWellKnownSid(sid, sid_type):
    bin_sid = byte_buffer(sid)
    return _IsWellKnownSid(ref(bin_sid), sid_type) != 0

################################################################################

_CheckTokenMembership = fun_fact(
    _adv.CheckTokenMembership, (
        BOOL,
        HANDLE,
        PVOID,
        PBOOL
        )
    )

def CheckTokenMembership(token_handle, sid_to_check):
    res = BOOL()
    sid = byte_buffer(sid_to_check)
    raise_on_zero(_CheckTokenMembership(token_handle, ref(sid), ref(res)))
    return res.value != 0

################################################################################

def running_as_admin():
    return CheckTokenMembership(
        None,
        CreateWellKnownSid(WinBuiltinAdministratorsSid)
        )

################################################################################

_OpenProcessToken = fun_fact(
    _adv.OpenProcessToken, (BOOL, HANDLE, DWORD, PHANDLE)
    )

def OpenProcessToken(proc_handle, desired_acc):
    token = KHANDLE()
    raise_on_zero(_OpenProcessToken(proc_handle, desired_acc, ref(token)))
    return token

################################################################################

_OpenThreadToken = fun_fact(
    _adv.OpenThreadToken, (BOOL, HANDLE, DWORD, BOOL, PHANDLE)
    )

def OpenThreadToken(thrd_handle, desired_acc, as_self):
    token = KHANDLE()
    raise_on_zero(
        _OpenThreadToken(thrd_handle, desired_acc, as_self, ref(token))
        )
    return token

################################################################################

def GetCurrentProcessToken():
    return HANDLE(-4)

def GetCurrentThreadToken():
    return HANDLE(-5)

def GetCurrentThreadEffectiveToken():
    return HANDLE(-6)

################################################################################

_DuplicateTokenEx = fun_fact(
    _adv.DuplicateTokenEx,
    (BOOL, HANDLE, DWORD, PSECURITY_ATTRIBUTES, INT, INT, PHANDLE)
    )

def DuplicateTokenEx(tok, acc, sattr, imp, typ):
    dup = KHANDLE()
    raise_on_zero(_DuplicateTokenEx(tok, acc, ref(sattr), imp, typ, ref(dup)))
    return dup

################################################################################

_AllocateLocallyUniqueId = fun_fact(
    _adv.AllocateLocallyUniqueId, (BOOL, PLUID)
    )

def AllocateLocallyUniqueId():
    luid = LUID()
    raise_on_zero(_AllocateLocallyUniqueId(ref(luid)))
    return int(luid)

################################################################################

class SID_AND_ATTRIBUTES(ctypes.Structure):
    _fields_ = (
        ("Sid", PVOID),
        ("Attributes", DWORD),
        )

class TOKEN_STATISTICS(ctypes.Structure):
    _fields_ = (
        ("TokenId", LUID),
        ("AuthenticationId", LUID),
        ("ExpirationTime", LARGE_INTEGER),
        ("TokenType", INT),
        ("ImpersonationLevel", INT),
        ("DynamicCharged", DWORD),
        ("DynamicAvailable", DWORD),
        ("GroupCount", DWORD),
        ("PrivilegeCount", DWORD),
        ("ModifiedId", LUID),
        )

################################################################################

_GetTokenInformation = fun_fact(
    _adv.GetTokenInformation, (BOOL, HANDLE, INT, PVOID, DWORD, PDWORD)
    )

def GetTokenInformation(hdl, cls):
    rlen = DWORD(256)
    while True:
        size = rlen.value
        buf = byte_buffer(size)
        if _GetTokenInformation(hdl, cls, buf, size, ref(rlen)):
            return buf.raw[:rlen.value]
        elif (err := GetLastError()) != ERROR_INSUFFICIENT_BUFFER:
            raise WinError(err)

################################################################################

_SetTokenInformation = fun_fact(
    _adv.SetTokenInformation, (BOOL, HANDLE, INT, PVOID, DWORD)
    )

def SetTokenInformation(hdl, cls, info):
    raise_on_zero(
        _SetTokenInformation(hdl, cls, ref(info), ctypes.sizeof(info))
        )

################################################################################

def _xform_saa(saa):
    return (ctypes.string_at(saa.Sid, GetLengthSid(saa.Sid)), saa.Attributes)

def get_token_user(hdl):
    saa = SID_AND_ATTRIBUTES.from_buffer_copy(
        GetTokenInformation(hdl, TokenUser)
        )
    return _xform_saa(saa)

################################################################################

def get_token_groups(hdl):
    buf = GetTokenInformation(hdl, TokenGroups)
    num_groups = DWORD.from_buffer_copy(buf).value
    class TOKEN_GROUPS(ctypes.Structure):
        _fields_ = (
            ("GroupCount", DWORD),
            ("Groups", SID_AND_ATTRIBUTES * num_groups),
            )
    tgroups = TOKEN_GROUPS.from_buffer_copy(buf)
    return (_xform_saa(saa) for saa in tgroups.Groups)

################################################################################

def get_token_privileges(hdl):
    buf = GetTokenInformation(hdl, TokenPrivileges)
    num_privs = DWORD.from_buffer_copy(buf).value
    class TOKEN_PRIVILEGES(ctypes.Structure):
        _fields_ = (
            ("PrivilegeCount", DWORD),
            ("Privileges", LUID_AND_ATTRIBUTES * num_privs)
            )
    tp = TOKEN_PRIVILEGES.from_buffer_copy(buf)
    return [(p.Luid, p.Attributes) for p in tp.Privileges]

################################################################################

def make_token_groups(sids_and_attrs):
    num_groups = len(sids_and_attrs)
    class TOKEN_GROUPS(ctypes.Structure):
        _fields_ = (
            ("GroupCount", DWORD),
            ("Groups", SID_AND_ATTRIBUTES * num_groups),
            ("sid_bufs", ctypes.py_object),
            )
    tgroups = TOKEN_GROUPS(num_groups, sid_bufs=[None] * num_groups)
    for i, (s, a) in enumerate(sids_and_attrs):
        tgroups.sid_bufs[i] = byte_buffer(s)
        tgroups.Groups[i].Sid = ctypes.addressof(tgroups.sid_bufs[i])
        tgroups.Groups[i].Attributes = a
    return tgroups

################################################################################

def running_as_system(consider_impersonation=False):
    token = (
        GetCurrentThreadEffectiveToken() if consider_impersonation
        else GetCurrentProcessToken()
        )
    sid, _ = get_token_user(token)
    return IsWellKnownSid(sid, WinLocalSystemSid)

################################################################################

def open_system_token_for_session(access, session):
    # Most likely the only process that can be used to retrieve the requested
    # token will be "winlogon.exe".
    if not (access & TOKEN_QUERY):
        access = access | TOKEN_QUERY
    for p in enum_processes():
        open_args = (PROCESS_QUERY_LIMITED_INFORMATION, False, p.pid)
        try:
            with OpenProcess(*open_args) as proc:
                token = OpenProcessToken(proc, access)
        except OSError:
            continue
        sid, _ = get_token_user(token)
        if IsWellKnownSid(sid, WinLocalSystemSid):
            buf = GetTokenInformation(token, TokenSessionId)
            if session == DWORD.from_buffer_copy(buf).value:
                return token
        CloseHandle(token)
    raise WinError(ERROR_NOT_FOUND)

################################################################################

def get_token_elevation_type(hdl):
    return int.from_bytes(
        GetTokenInformation(hdl, TokenElevationType),
        "little"
        )

################################################################################

def is_elevated_via_uac():
    ele_type = get_token_elevation_type(GetCurrentProcessToken())
    return ele_type == TokenElevationTypeFull

################################################################################

_CreateProcessAsUser = fun_fact(
    _adv.CreateProcessAsUserW, (
        BOOL,
        HANDLE,
        PWSTR,
        PWSTR,
        PSECURITY_ATTRIBUTES,
        PSECURITY_ATTRIBUTES,
        BOOL,
        DWORD,
        PVOID,
        PWSTR,
        PSTARTUPINFO,
        PPROCESS_INFORMATION,
        )
    )

def CreateProcessAsUser(
        token,
        app_name,
        cmd_line,
        proc_attr,
        thread_attr,
        inherit,
        cflags,
        env,
        curdir,
        startup_info
        ):
    proc_info = PROCESS_INFORMATION()
    raise_on_zero(
        _CreateProcessAsUser(
            token,
            app_name,
            cmd_line,
            ref(proc_attr) if proc_attr is not None else None,
            ref(thread_attr) if thread_attr is not None else None,
            inherit,
            cflags,
            env,
            curdir,
            ref(startup_info),
            ref(proc_info)
            )
        )
    return proc_info

################################################################################

def create_process_as_user(
        token,
        arglist,
        cflags=0,
        startup_info=None,
        inherit=False,
        env=None,
        curdir=None,
        proc_attr=None,
        thread_attr=None,
        ):
    if startup_info is None:
        startup_info = STARTUPINFO()
    return CreateProcessAsUser(
        token,
        None,
        cmdline_from_args(arglist),
        proc_attr,
        thread_attr,
        inherit,
        cflags,
        env,
        curdir,
        startup_info
        )

################################################################################

_LookupPrivilegeValue = fun_fact(
    _adv.LookupPrivilegeValueW, (BOOL, PWSTR, PWSTR, PLUID)
    )

def LookupPrivilegeValue(sys_name, name):
    luid = LUID()
    raise_on_zero(_LookupPrivilegeValue(sys_name, name, ref(luid)))
    return luid

################################################################################

class LUID_AND_ATTRIBUTES(ctypes.Structure):
    _fields_ = (
        ("Luid", LUID),
        ("Attributes", DWORD)
        )

def AdjustTokenPrivileges(token, luids_and_attributes, disable_all=False):
    num_la = len(luids_and_attributes)
    if not num_la:
        return

    class TOKEN_PRIVILEGES(ctypes.Structure):
        _fields_ = (
            ("PrivilegeCount", DWORD),
            ("Privileges", LUID_AND_ATTRIBUTES * num_la)
            )
    PTOKEN_PRIVILEGES = POINTER(TOKEN_PRIVILEGES)
    _AdjustTokenPrivileges = fun_fact(
        _adv.AdjustTokenPrivileges, (
            BOOL,
            HANDLE,
            BOOL,
            PTOKEN_PRIVILEGES,
            DWORD,
            PTOKEN_PRIVILEGES,
            PDWORD,
            )
        )
    privs = TOKEN_PRIVILEGES()
    privs.PrivilegeCount = num_la
    for n, la in enumerate(luids_and_attributes):
        privs.Privileges[n].Luid = la.Luid
        privs.Privileges[n].Attributes = la.Attributes

    raise_on_zero(
        _AdjustTokenPrivileges(
            token,
            disable_all,
            ref(privs),
            0,
            None,
            None
            )
        )
    if (err := GetLastError()) == ERROR_NOT_ALL_ASSIGNED:
        raise WinError(err)

################################################################################

def enable_token_privileges(token, privileges):
    laa = [
        LUID_AND_ATTRIBUTES(LUID(p), SE_PRIVILEGE_ENABLED) for p in privileges
    ]
    AdjustTokenPrivileges(token, laa)

################################################################################

def enable_privileges(privileges):
    with OpenProcessToken(GetCurrentProcess(), MAXIMUM_ALLOWED) as tok:
        enable_token_privileges(tok, privileges)

################################################################################

_LookupAccountSid = fun_fact(
    _adv.LookupAccountSidW,
    (BOOL, PWSTR, PVOID, PWSTR, PDWORD, PWSTR, PDWORD, PDWORD)
    )

def LookupAccountSid(sid, system_name=None):
    name_size = DWORD(0)
    domain_size = DWORD(0)
    sid_use = DWORD()
    ok = _LookupAccountSid(
        system_name,
        sid,
        None,
        ref(name_size),
        None,
        ref(domain_size),
        ref(sid_use)
        )
    err = GetLastError()
    if ok:
        raise AssertionError("logic error in LookupAccountSid")
    if err != ERROR_INSUFFICIENT_BUFFER:
        raise WinError(err)

    name = string_buffer(name_size.value)
    domain = string_buffer(domain_size.value)
    raise_on_zero(
        _LookupAccountSid(
            system_name,
            sid,
            name,
            ref(name_size),
            domain,
            ref(domain_size),
            ref(sid_use)
            )
        )
    return name.value, domain.value, sid_use.value

################################################################################

_SetThreadToken = fun_fact(_adv.SetThreadToken, (BOOL, PHANDLE, HANDLE))

def SetThreadToken(tok, thrd=None):
    pht = None if thrd is None else ref(thrd)
    raise_on_zero(_SetThreadToken(pht, tok))

################################################################################

class ACL(ctypes.Structure):
    _fields_ = (
        ("AclRevision", BYTE),
        ("Sbz1", BYTE),
        ("AclSize", WORD),
        ("AceCount", WORD),
        ("Sbz2", WORD),
        )
PACL = POINTER(ACL)
PPACL = POINTER(PACL)

class ACE_HEADER(ctypes.Structure):
    _fields_ = (
        ("AceType", BYTE),
        ("AceFlags", BYTE),
        ("AceSize", WORD),
        )

class ACE(ctypes.Structure):
    _fields_ = (
        ("Header", ACE_HEADER),
        ("Mask", DWORD),

        # first DWORD of SID, remaining bytes of the SID are stored in
        # contiguous memory after SidStart
        ("SidStart", DWORD),
        )
PACE = POINTER(ACE)
PPACE = POINTER(PACE)

class ACL_SIZE_INFORMATION(ctypes.Structure):
    _fields_ = (
        ("AceCount", DWORD),
        ("AclBytesInUse", DWORD),
        ("AclBytesFree", DWORD),
        )

################################################################################

_GetAce = fun_fact(
    _adv.GetAce,
    (BOOL, PACL, DWORD, PPACE)
    )

def GetAce(pacl, idx):
    pace = PACE()
    raise_on_zero(_GetAce(pacl, idx, ref(pace)))
    return pace

################################################################################

_GetSecurityDescriptorDacl = fun_fact(
    _adv.GetSecurityDescriptorDacl,
    (BOOL, PVOID, PBOOL, PPACL, PBOOL)
    )

def GetSecurityDescriptorDacl(sd):
    present = BOOL()
    pacl = PACL()
    defaulted = BOOL()
    raise_on_zero(
        _GetSecurityDescriptorDacl(
            sd,
            ref(present),
            ref(pacl),
            ref(defaulted)
            )
        )
    return present, pacl, defaulted

################################################################################

_GetSecurityDescriptorOwner = fun_fact(
    _adv.GetSecurityDescriptorOwner,
    (BOOL, PVOID, PVOID, PBOOL)
    )

def GetSecurityDescriptorOwner(sd):
    psid = PVOID()
    defaulted = BOOL()
    raise_on_zero(
        _GetSecurityDescriptorOwner(
            sd,
            ref(psid),
            ref(defaulted)
            )
        )
    return ctypes.string_at(psid, GetLengthSid(psid)), bool(defaulted)

################################################################################

_GetSecurityDescriptorGroup = fun_fact(
    _adv.GetSecurityDescriptorGroup,
    (BOOL, PVOID, PVOID, PBOOL)
    )

def GetSecurityDescriptorGroup(sd):
    psid = PVOID()
    defaulted = BOOL()
    raise_on_zero(
        _GetSecurityDescriptorGroup(
            sd,
            ref(psid),
            ref(defaulted)
            )
        )
    return ctypes.string_at(psid, GetLengthSid(psid)), defaulted

################################################################################

GetSecurityDescriptorLength = fun_fact(
    _adv.GetSecurityDescriptorLength, (DWORD, PVOID)
    )

################################################################################

_GetNamedSecurityInfo = fun_fact(
    _adv.GetNamedSecurityInfoW,
    (DWORD, PWSTR, DWORD, DWORD, PPVOID, PPVOID, PPVOID, PPVOID, PPVOID)
    )

NEARLY_ALL_SECURITY_INFORMATION = (
    OWNER_SECURITY_INFORMATION |
    GROUP_SECURITY_INFORMATION |
    DACL_SECURITY_INFORMATION
    )

def GetNamedSecurityInfo(name, otype, req_info=NEARLY_ALL_SECURITY_INFORMATION):
    pOwner = PVOID()
    pGroup = PVOID()
    pDacl = PVOID()
    pSacl = PVOID()
    pSD = PVOID()
    raise_on_err(
        _GetNamedSecurityInfo(
            name,
            otype,
            req_info,
            ref(pOwner),
            ref(pGroup),
            ref(pDacl),
            ref(pSacl),
            ref(pSD)
            )
        )
    try:
        return ctypes.string_at(pSD.value, GetSecurityDescriptorLength(pSD))
    finally:
        LocalFree(pSD)

################################################################################

_SetNamedSecurityInfo = fun_fact(
    _adv.SetNamedSecurityInfoW,
    (DWORD, PWSTR, DWORD, DWORD, PVOID, PVOID, PVOID, PVOID)
    )

def SetNamedSecurityInfo(
        name,
        otype,
        *,
        owner=None,
        group=None,
        dacl=None,
        sacl=None
        ):
    set_info = 0
    if owner is not None:
        set_info |= OWNER_SECURITY_INFORMATION
    if group is not None:
        set_info |= GROUP_SECURITY_INFORMATION
    if dacl is not None:
        set_info |= DACL_SECURITY_INFORMATION
    if sacl is not None:
        set_info |= SACL_SECURITY_INFORMATION
    raise_on_err(
        _SetNamedSecurityInfo(
            name,
            otype,
            set_info,
            owner,
            group,
            dacl,
            sacl
            )
        )

################################################################################

_CloseServiceHandle = fun_fact(_adv.CloseServiceHandle, (BOOL, HANDLE))

def CloseServiceHandle(handle):
    raise_on_zero(_CloseServiceHandle(handle))

################################################################################

class SC_HANDLE(
        ScdToBeClosed,
        HANDLE,
        close_func=CloseServiceHandle,
        invalid=0
        ):
    pass

################################################################################

_OpenSCManager = fun_fact(
    _adv.OpenSCManagerW, (HANDLE, PWSTR, PWSTR, DWORD)
    )

def OpenSCManager(machine_name, database_name, desired_acc):
    res = SC_HANDLE(_OpenSCManager(machine_name, database_name, desired_acc))
    res.raise_on_invalid()
    return res

################################################################################

_OpenService = fun_fact(
    _adv.OpenServiceW, (HANDLE, HANDLE, PWSTR, DWORD)
    )

def OpenService(scm, name, desired_acc):
    res = SC_HANDLE(_OpenService(scm, name, desired_acc))
    res.raise_on_invalid()
    return res

################################################################################

_CreateService = fun_fact(
    _adv.CreateServiceW, (
        HANDLE,
        HANDLE,
        PWSTR,
        PWSTR,
        DWORD,
        DWORD,
        DWORD,
        DWORD,
        PWSTR,
        PWSTR,
        PDWORD,
        PWSTR,
        PWSTR,
        PWSTR,
        )
    )

def CreateService(
        scm,
        service_name,
        display_name,
        desired_acc,
        service_type,
        start_type,
        error_control,
        binary_path_name,
        load_order_group=None,
        dependencies=None,
        service_start_name=None,
        password=None
        ):
    if dependencies is not None:
        dependencies = string_buffer("\x00".join(dependencies))
    res = SC_HANDLE(
        _CreateService(
            scm,
            service_name,
            display_name,
            desired_acc,
            service_type,
            start_type,
            error_control,
            binary_path_name,
            load_order_group,
            None,
            dependencies,
            service_start_name,
            password
            )
        )
    res.raise_on_invalid()
    return res

################################################################################

_StartService = fun_fact(
    _adv.StartServiceW, (
        BOOL,
        HANDLE,
        DWORD,
        PVOID
        )
    )

def StartService(handle, arglist):
    aa = ArgcArgvFromArgs(arglist)
    raise_on_zero(_StartService(handle, aa.argc, aa.argv))

################################################################################

class SERVICE_STATUS(ctypes.Structure):
    _fields_ = (
        ("dwServiceType", DWORD),
        ("dwCurrentState", DWORD),
        ("dwControlsAccepted", DWORD),
        ("dwWin32ExitCode", DWORD),
        ("dwServiceSpecificExitCode", DWORD),
        ("dwCheckPoint", DWORD),
        ("dwWaitHint", DWORD),
        )
PSERVICE_STATUS = POINTER(SERVICE_STATUS)

_ControlService = fun_fact(
    _adv.ControlService, (
        BOOL,
        HANDLE,
        DWORD,
        PSERVICE_STATUS
        )
    )

def ControlService(service, control):
    status = SERVICE_STATUS()
    raise_on_zero(_ControlService(service, control, ref(status)))
    return status

################################################################################

_DeleteService = fun_fact(_adv.DeleteService, (BOOL, HANDLE))

def DeleteService(service):
    raise_on_zero(_DeleteService(service))

################################################################################

class SERVICE_STATUS_PROCESS(ctypes.Structure):
    _fields_ = (
        ("dwServiceType", DWORD),
        ("dwCurrentState", DWORD),
        ("dwControlsAccepted", DWORD),
        ("dwWin32ExitCode", DWORD),
        ("dwServiceSpecificExitCode", DWORD),
        ("dwCheckPoint", DWORD),
        ("dwWaitHint", DWORD),
        ("dwProcessId", DWORD),
        ("dwServiceFlags", DWORD),
        )

_QueryServiceStatusEx = fun_fact(
    _adv.QueryServiceStatusEx, (
        BOOL,
        HANDLE,
        INT,
        POINTER(SERVICE_STATUS_PROCESS),
        DWORD,
        PDWORD
        )
    )

def QueryServiceStatusEx(service):
    status = SERVICE_STATUS_PROCESS()
    needed = DWORD()
    raise_on_zero(
        _QueryServiceStatusEx(
            service,
            SC_STATUS_PROCESS_INFO,
            ref(status),
            ctypes.sizeof(status),
            ref(needed)
            )
        )
    return status

################################################################################

class ENUM_SERVICE_STATUS_PROCESS(ctypes.Structure):
    _fields_ = (
        ("lpServiceName", PWSTR),
        ("lpDisplayName", PWSTR),
        ("ServiceStatusProcess", SERVICE_STATUS_PROCESS),
        )

_EnumServicesStatusEx = fun_fact(
    _adv.EnumServicesStatusExW, (
        BOOL,
        HANDLE,
        INT,
        DWORD,
        DWORD,
        PVOID,
        DWORD,
        PDWORD,
        PDWORD,
        PDWORD,
        PWSTR
        )
    )

def EnumServicesStatusEx(scm, stype, sstate, group_name=None):
    stat_size = ctypes.sizeof(ENUM_SERVICE_STATUS_PROCESS)
    buf = byte_buffer(16 * 1024)
    needed = DWORD()
    num_ret = DWORD()
    resume = DWORD()
    result = []
    success = False

    while not success:
        success = _EnumServicesStatusEx(
            scm,
            SC_ENUM_PROCESS_INFO,
            stype,
            sstate,
            buf,
            ctypes.sizeof(buf),
            ref(needed),
            ref(num_ret),
            ref(resume),
            group_name
            )
        raise_if(not success and GetLastError() != ERROR_MORE_DATA)

        for offs in range(0, num_ret.value * stat_size, stat_size):
            status = ENUM_SERVICE_STATUS_PROCESS.from_buffer(buf, offs)
            result.append(ns_from_struct(status))

        if success:
            break
        buf = byte_buffer(needed.value)

    return result

################################################################################

class QUERY_SERVICE_CONFIG(ctypes.Structure):
    _fields_ = (
        ("dwServiceType", DWORD),
        ("dwStartType", DWORD),
        ("dwErrorControl", DWORD),
        ("lpBinaryPathName", PWSTR),
        ("lpLoadOrderGroup", PWSTR),
        ("dwTagId", DWORD),
        ("lpDependencies", PWSTR),
        ("lpServiceStartName", PWSTR),
        ("lpDisplayName", PWSTR),
        )
PQUERY_SERVICE_CONFIG = POINTER(QUERY_SERVICE_CONFIG)

_QueryServiceConfig = fun_fact(
    _adv.QueryServiceConfigW, (
        BOOL,
        HANDLE,
        PQUERY_SERVICE_CONFIG,
        DWORD,
        PDWORD,
        )
    )

def QueryServiceConfig(svc):
    needed = DWORD()
    ok = _QueryServiceConfig(svc, None, 0, ref(needed))
    err = GetLastError()
    if ok:
        raise AssertionError("logic error in QueryServiceConfig")
    if err != ERROR_INSUFFICIENT_BUFFER:
        raise WinError(err)
    buf = byte_buffer(needed.value)
    pqsc = ctypes.cast(buf, PQUERY_SERVICE_CONFIG)
    raise_on_zero(_QueryServiceConfig(svc, pqsc, needed.value, ref(needed)))
    return ns_from_struct(pqsc.contents)

################################################################################

SERVICE_MAIN_FUNCTION = ctypes.WINFUNCTYPE(None, DWORD, PPWSTR)

class SERVICE_TABLE_ENTRY(ctypes.Structure):
    _fields_ = (
        ("lpServiceName", PWSTR),
        ("lpServiceProc", SERVICE_MAIN_FUNCTION),
        )
PSERVICE_TABLE_ENTRY = POINTER(SERVICE_TABLE_ENTRY)

_StartServiceCtrlDispatcher = fun_fact(
    _adv.StartServiceCtrlDispatcherW, (BOOL, PSERVICE_TABLE_ENTRY)
    )

def StartServiceCtrlDispatcher(table):
    raise_on_zero(_StartServiceCtrlDispatcher(ref(table[0])))

################################################################################

HANDLER_FUNCTION = ctypes.WINFUNCTYPE(None, DWORD)

_RegisterServiceCtrlHandler = fun_fact(
    _adv.RegisterServiceCtrlHandlerW, (HANDLE, PWSTR, HANDLER_FUNCTION)
    )

def RegisterServiceCtrlHandler(name, handler):
    res = _RegisterServiceCtrlHandler(name, handler)
    raise_on_zero(res)
    return res

################################################################################

_SetServiceStatus = fun_fact(
    _adv.SetServiceStatus, (BOOL, HANDLE, PSERVICE_STATUS)
    )

def SetServiceStatus(hdl, status):
    raise_on_zero(_SetServiceStatus(hdl, ref(status)))

################################################################################

class CREDENTIAL_ATTRIBUTE(ctypes.Structure):
    _fields_ = (
        ("Keyword", PWSTR),
        ("Flags", DWORD),
        ("ValueSize", DWORD),
        ("Value", PBYTE),
        )
PCREDENTIAL_ATTRIBUTE = POINTER(CREDENTIAL_ATTRIBUTE)

class CREDENTIAL(ctypes.Structure):
    _fields_ = (
        ("Flags", DWORD),
        ("Type", DWORD),
        ("TargetName", PWSTR),
        ("Comment", PWSTR),
        ("LastWritten", FILETIME),
        ("CredentialBlobSize", DWORD),
        ("CredentialBlob", PBYTE),
        ("Persist", DWORD),
        ("AttributeCount", DWORD),
        ("Attributes", PCREDENTIAL_ATTRIBUTE),
        ("TargetAlias", PWSTR),
        ("UserName", PWSTR),
        )
PCREDENTIAL = POINTER(CREDENTIAL)
PPCREDENTIAL = POINTER(PCREDENTIAL)
PPPCREDENTIAL = POINTER(PPCREDENTIAL)

_CredRead = fun_fact(
    _adv.CredReadW, (
        BOOL,
        PWSTR,
        DWORD,
        DWORD,
        PPCREDENTIAL
        )
    )
_CredEnumerate = fun_fact(
    _adv.CredEnumerateW, (
        BOOL,
        PWSTR,
        DWORD,
        PDWORD,
        PPPCREDENTIAL
        )
    )
_CredWrite = fun_fact(_adv.CredWriteW, (BOOL, PCREDENTIAL, DWORD))
_CredFree = fun_fact(_adv.CredFree, (None, PVOID))

################################################################################

def _ns_from_cred(cred):
    def a2ns(attr):
        return _namespace(
            Keyword=attr.Keyword,
            Flags=attr.Flags,
            Value=ctypes.string_at(attr.Value, attr.ValueSize),
            )
    attr = tuple(a2ns(cred.Attributes[n]) for n in range(cred.AttributeCount))
    blob = ctypes.string_at(cred.CredentialBlob, cred.CredentialBlobSize)
    return _namespace(
        TargetName=cred.TargetName,
        UserName=cred.UserName,
        CredentialBlob=blob,
        Attributes=attr,
        Flags=cred.Flags,
        Type=cred.Type,
        Comment=cred.Comment,
        LastWritten=cred.LastWritten,
        Persist=cred.Persist,
        TargetAlias=cred.TargetAlias,
        )

################################################################################

def CredRead(target_name, typ=CRED_TYPE_GENERIC, flags=0):
    ptr = PCREDENTIAL()
    try:
        raise_on_zero(_CredRead(target_name, typ, flags, ref(ptr)))
        return _ns_from_cred(ptr.contents)
    finally:
        _CredFree(ptr)

################################################################################

def CredEnumerate(filter=None, flags=0):
    pptr = PPCREDENTIAL()
    cnt = DWORD()
    try:
        raise_on_zero(_CredEnumerate(filter, flags, ref(cnt), ref(pptr)))
        return tuple(_ns_from_cred(pptr[n].contents) for n in range(cnt.value))
    finally:
        _CredFree(pptr)

################################################################################

def CredWrite(credential, flags=0):

    # ============================== std fields ================================

    std_fields = (
        "Flags",
        "Type",
        "TargetName",
        "Comment",
        "LastWritten",
        "Persist",
        "TargetAlias",
        "UserName",
        )
    cred = CREDENTIAL()
    for f in std_fields:
        if (val := getattr(credential, f, None)) is not None:
            setattr(cred, f, val)

    # ================================= blob ===================================

    if (val := getattr(credential, "CredentialBlob", None)) is not None:
        cred.CredentialBlobSize = len(val)
        cred.CredentialBlob = (BYTE * len(val))(*tuple(map(int, val)))

    # ============================== attributes ================================

    if ns_attr := getattr(credential, "Attributes", None):
        attr = (CREDENTIAL_ATTRIBUTE * len(ns_attr))()
        for i, a in enumerate(ns_attr):
            for f in ("Keyword", "Flags"):
                if (val := getattr(a, f, None)) is not None:
                    setattr(attr[i], f, val)
            if val := getattr(a, "Value", None):
                attr[i].ValueSize = len(val)
                attr[i].Value = (BYTE * len(val))(*tuple(map(int, val)))
        cred.AttributeCount = len(ns_attr)
        cred.Attributes = ctypes.cast(attr, PCREDENTIAL_ATTRIBUTE)

    raise_on_zero(_CredWrite(ref(cred), flags))

################################################################################

_CloseEventLog = fun_fact(_adv.CloseEventLog, (BOOL, HANDLE))

def CloseEventLog(hdl):
    raise_on_zero(_CloseEventLog(hdl))

################################################################################

class EHANDLE(ScdToBeClosed, HANDLE, close_func=CloseEventLog, invalid=0):
    pass

################################################################################

_OpenEventLog = fun_fact(
    _adv.OpenEventLogW, (HANDLE, PWSTR, PWSTR)
    )

def OpenEventLog(source, server=None):
    hdl = EHANDLE(_OpenEventLog(server, source))
    hdl.raise_on_invalid()
    return hdl

################################################################################

class EVENTLOGRECORD(ctypes.Structure):
    _fields_ = (
        ("Length", DWORD),
        ("Reserved", DWORD),
        ("RecordNumber", DWORD),
        ("TimeGenerated", DWORD),
        ("TimeWritten", DWORD),
        ("EventID", DWORD),
        ("EventType", WORD),
        ("NumStrings", WORD),
        ("EventCategory", WORD),
        ("ReservedFlags", WORD),
        ("ClosingRecordNumber", DWORD),
        ("StringOffset", DWORD),
        ("UserSidLength", DWORD),
        ("UserSidOffset", DWORD),
        ("DataLength", DWORD),
        ("DataOffset", DWORD),
        #
        # followed by:
        #
        # WCHAR SourceName[]
        # WCHAR Computername[]
        # SID   UserSid
        # WCHAR Strings[]
        # BYTE  Data[]
        # CHAR  Pad[]
        # DWORD Length;
        )

PEVENTLOGRECORD = POINTER(EVENTLOGRECORD)

################################################################################

def _evt_from_buf(buf, offs):
    addr = ctypes.addressof(buf) + offs
    elr = EVENTLOGRECORD.from_buffer(buf, offs)

    str_ins = []
    if elr.NumStrings:
        stroffs = elr.StringOffset
        for _ in range(elr.NumStrings):
            nxt = ctypes.wstring_at(addr + stroffs)
            stroffs += (len(nxt) + 1) * WCHAR_SIZE
            str_ins.append(nxt)
    sid = ""
    if elr.UserSidLength:
        sid = ConvertSidToStringSid(
            ctypes.string_at(addr + elr.UserSidOffset, elr.UserSidLength)
            )
    data = ctypes.string_at(addr + elr.DataOffset, elr.DataLength)
    p_str = addr + ctypes.sizeof(EVENTLOGRECORD)
    src_name = ctypes.wstring_at(p_str)
    p_str += (len(src_name) + 1) * WCHAR_SIZE
    computer_name = ctypes.wstring_at(p_str)
    ltz = get_local_tzinfo()
    tgen = _dt.fromtimestamp(elr.TimeGenerated, tz=_UTC).astimezone(ltz)
    twri = _dt.fromtimestamp(elr.TimeWritten, tz=_UTC).astimezone(ltz)

    return elr.Length, _namespace(
        ClosingRecordNumber=elr.ClosingRecordNumber,
        ComputerName=computer_name,
        Data=data,
        EventCategory=elr.EventCategory,
        EventID=elr.EventID,
        EventType=elr.EventType,
        RecordNumber=elr.RecordNumber,
        Reserved=elr.Reserved,
        ReservedFlags=elr.ReservedFlags,
        Sid=sid,
        SourceName=src_name,
        StringInserts=str_ins,
        TimeGenerated=tgen,
        TimeWritten=twri,
        )

################################################################################

_ReadEventLog = fun_fact(
    _adv.ReadEventLogW, (
        BOOL,
        HANDLE,
        DWORD,
        DWORD,
        PVOID,
        DWORD,
        PDWORD,
        PDWORD,
        )
    )

def ReadEventLog(hdl, flags=None, offs=0, size=16384):
    if flags is None:
        flags = EVENTLOG_SEQUENTIAL_READ | EVENTLOG_BACKWARDS_READ
    want = DWORD(size)
    while True:
        got = DWORD()
        buf = byte_buffer(want.value)
        ok = _ReadEventLog(hdl, flags, offs, buf, want, ref(got), ref(want))
        got = got.value
        if not ok:
            err = GetLastError()
            if err == ERROR_HANDLE_EOF:
                got = 0
                break
            elif err == ERROR_INSUFFICIENT_BUFFER:
                continue
            else:
                raise WinError(err)
        else:
            break

    res = []
    offs = 0
    while offs < got:
        elen, evt = _evt_from_buf(buf, offs)
        res.append(evt)
        offs += elen
    return res

################################################################################

def enum_event_log(name, flags=None, server=None):
    with OpenEventLog(name, server) as log:
        while some_events := ReadEventLog(log, flags):
            yield from some_events

################################################################################

_EncryptFile = fun_fact(_adv.EncryptFileW, (BOOL, PWSTR))

def EncryptFile(file_or_dir_name):
    raise_on_zero(_EncryptFile(file_or_dir_name))

################################################################################

_DecryptFile = fun_fact(_adv.DecryptFileW, (BOOL, PWSTR, DWORD))

def DecryptFile(file_or_dir_name):
    raise_on_zero(_DecryptFile(file_or_dir_name, 0))

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

# This sample was inspired by Charles Petzold's 'keyview' sample from his
# 'Programming Windows' book (https://www.charlespetzold.com/pw5/index.html)

import dataclasses
from ctwin32 import (
    user,
    gdi,
    wndcls,
    WM_CREATE,
    WM_PAINT,
    WM_DESTROY,
    WM_SIZE,
    WM_KEYFIRST,
    WM_KEYLAST,
    WM_CHAR,
    WM_SYSCHAR,
    WM_DEADCHAR,
    WM_SYSDEADCHAR,
    DEFAULT_CHARSET,
    SYSTEM_FONT,
    SWP_NOMOVE,
    SM_CYMAXIMIZED,
    TRANSPARENT,
    )
from ctwin32.wtypes import RECT, LOGFONT

################################################################################

@dataclasses.dataclass
class HistoryEntry:
    msg: int = 0
    wp: int = 0
    lp: int = 0

################################################################################

class KeyViewWnd(wndcls.SimpleWnd):

    CHAR_MSG = (WM_CHAR, WM_SYSCHAR, WM_DEADCHAR, WM_SYSDEADCHAR)

    def __init__(self, wc_params):
        self.heading = (
            " message        key     char              rep.  "
            "scan   ext. alt   prev  now"
            )
        self.underline = "".join(
                (" " if c == " " else "_") for c in self.heading
                )
        self.msg_name = [
            "WM_KEYDOWN",
            "WM_KEYUP",
            "WM_CHAR",
            "WM_DEADCHAR",
            "WM_SYSKEYDOWN",
            "WM_SYSKEYUP",
            "WM_SYSCHAR",
            "WM_SYSDEADCHAR"
            ]
        self.font = None
        self.char_height = 0
        self.num_lines = 0
        self.scroll_rect = RECT()
        self.max_lines = 0

        # have to call __init__ of base class AFTER preparing members that
        # might get accessed by on_message
        super().__init__(wc_params)

    ############################################################################

    def on_create(self):
        hdc = self.get_dc()
        lf = LOGFONT()
        lf.lfCharSet = DEFAULT_CHARSET
        lf.lfFaceName = "consolas"
        lf.lfHeight = 16
        self.font = gdi.CreateFontIndirect(lf)
        gdi.SelectObject(hdc, self.font)
        tm = gdi.GetTextMetrics(hdc)
        gdi.SelectObject(hdc, gdi.GetStockObject(SYSTEM_FONT))
        self.release_dc(hdc)

        self.char_height = tm.tmHeight
        size = (len(self.underline) + 4) * tm.tmAveCharWidth
        self.set_pos(None, 0, 0, size, size, SWP_NOMOVE)

        self.max_lines = (
            user.GetSystemMetrics(SM_CYMAXIMIZED) // self.char_height
            )
        self.history = [None] * self.max_lines

        return 0

    ############################################################################

    def on_paint(self):
        hdc, ps = self.begin_paint()
        prev_font = gdi.SelectObject(hdc, self.font)

        gdi.SetBkMode(hdc, TRANSPARENT)
        gdi.TextOut(hdc, 0, 0, self.heading)
        gdi.TextOut(hdc, 0, 0, self.underline)


        for i in range(min(self.num_lines, self.y_lines)):
            then = self.history[i]
            rep = then.lp & 0xffff
            scan = (then.lp >> 16) & 0xff
            ext = "yes" if 0x01000000 & then.lp else "no"
            alt = "yes" if 0x20000000 & then.lp else "no"
            prev = "down" if 0x40000000 & then.lp else "up"
            now = "up" if 0x80000000 & then.lp else "down"
            name = self.msg_name[then.msg - WM_KEYFIRST]

            if then.msg in self.CHAR_MSG:
                middle = f"          {then.wp:#06x}      {chr(then.wp)} "
            else:
                kn = user.GetKeyNameText(then.lp, True)
                middle = f" {then.wp:3} {kn:<18} "
            end = f"{rep:6}  {scan:4}     {ext:>3} {alt:>4} {prev:>6} {now:>4}"
            line = f" {name:<13}{middle}{end}"
            gdi.TextOut(hdc, 0, (self.y_lines - i) * self.char_height, line)

        gdi.SelectObject(hdc, prev_font)
        self.end_paint(ps)
        return 0

    ############################################################################

    def on_message(self, msg, wp, lp):

        if msg == WM_CREATE:
            return self.on_create()

        elif msg == WM_SIZE:
            self.y_lines = ((lp >> 16) & 0xffff) // self.char_height - 1
            self.scroll_rect.right  = lp & 0xffff
            self.scroll_rect.top    = self.char_height
            self.scroll_rect.bottom = self.char_height * (self.y_lines + 1)
            self.invalidate_rect(erase=True)
            return 0

        elif WM_KEYFIRST <= msg < WM_KEYLAST:
            self.history.pop()
            self.history.insert(0, HistoryEntry(msg, wp, lp))

            self.num_lines = min(self.num_lines + 1, self.max_lines)
            rc = self.scroll_rect
            user.ScrollWindow(self.hwnd, 0, -self.char_height, rc, rc)
            return self.def_win_proc(msg, wp, lp)

        elif msg == WM_PAINT:
            return self.on_paint()

        elif msg == WM_DESTROY:
            gdi.DeleteObject(self.font)
            user.PostQuitMessage(0)
            return 0

        else:
            return self.def_win_proc(msg, wp, lp)

################################################################################

if __name__ == "__main__":

    icon = wndcls.load_ctwin32_ico()
    wnd = KeyViewWnd(wndcls.WndCreateParams("keyview", icon))
    wnd.show()

    while msg := user.GetMessage():
        user.TranslateMessage(msg)
        user.DispatchMessage(msg)

################################################################################



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

import sys
from ctwin32 import advapi, shell
from ctwin32.setupapi import (
    get_non_present_info_set,
    enum_info_set,
    remove_non_present_devices,
    SetupDiGetDeviceInstanceId,
    desc_from_info_set,
    )

################################################################################

def print_non_present():
    print("non present devices:")
    for iset, ddat in enum_info_set(get_non_present_info_set()):
        desc = desc_from_info_set(iset, ddat)
        iid = SetupDiGetDeviceInstanceId(iset, ddat)
        print(f"{desc:40s} {iid}")

################################################################################

do_remove = len(sys.argv) > 1 and sys.argv[1] == "-r"
have_to_elevate = do_remove and not advapi.running_as_admin()

if have_to_elevate:
    shell.elevate(f'"{sys.executable}"', f'"{__file__}"', "-r")
else:
    print_non_present()
    if do_remove:
        print("\nremoving non present devices ...\n")
        remove_non_present_devices()
        print_non_present()



################################################################################
#
# Copyright 2021-2025 Rocco Matano
# SPDX-License-Identifier: MIT
#
################################################################################

# This sample demonstrates how to extract icons from *.ico files and from
# executable files (*.exe, *.dll etc.). It prints the extracted data in the
# format that is used in ctwin32.wndcls to store such data in python code
# (lzma-compressed and b85-encoded). You can compare the output of running
# 'python extract_ico.py ..\doc\images\ctwin32.ico 1' with the definition of
# ctwin32.wndcls._ctwin32_icon.

################################################################################

import sys
from types import SimpleNamespace
from ctwin32 import (
    ctypes,
    kernel,
    wndcls,
    LOAD_LIBRARY_AS_IMAGE_RESOURCE,
    LOAD_LIBRARY_AS_DATAFILE,
    RT_GROUP_ICON,
    RT_ICON,
    )
from ctwin32.wtypes import BYTE, WORD, DWORD

################################################################################

# https://docs.microsoft.com/en-us/previous-versions/ms997538(v=msdn.10)

class ICONDIR_HEADER(ctypes.Structure):
    _fields_ = (
        ("reserved", WORD),
        ("type", WORD),
        ("count", WORD),
        )

# ico file
class ICONDIRENTRY(ctypes.Structure):
    _fields_ = (
        ("width", BYTE),
        ("height", BYTE),
        ("colors", BYTE),
        ("reserved", BYTE),
        ("planes", WORD),
        ("bitcount", WORD),
        ("bytesinres", DWORD),
        ("imgoffest", DWORD),
        )

# icon resource
class GRPICONDIRENTRY(ctypes.Structure):
    _pack_ = 2
    _fields_ = (
        ("width", BYTE),
        ("height", BYTE),
        ("colors", BYTE),
        ("reserved", BYTE),
        ("planes", WORD),
        ("bitcount", WORD),
        ("bytesinres", DWORD),
        ("ico_id", WORD),
        )

################################################################################

def print_lzma_b85(data):
    b85 = wndcls.to_lzb85(data)
    llen = 72
    while len(b85) > llen:
        print(b85[:llen])
        b85 = b85[llen:]
    print(b85)

################################################################################

def get_ico_file_info(filename):
    with open(filename, "rb") as file:
        dta = file.read(ctypes.sizeof(ICONDIR_HEADER))
        idh = ICONDIR_HEADER.from_buffer_copy(dta)
        ICON_ENTRIES = ICONDIRENTRY * idh.count
        dta = file.read(ctypes.sizeof(ICON_ENTRIES))
        entries = ICON_ENTRIES.from_buffer_copy(dta)
        image_info = [
            SimpleNamespace(
                width=e.width if e.width else 256,
                height=e.height if e.height else 256,
                colors=e.colors if e.colors else 1 << min(e.bitcount, 24),
                size=e.bytesinres,
                offset=e.imgoffest,
                )
            for e in entries
            ]
        return [(0, image_info)]

################################################################################

def extract_from_ico(filename, fmt):
    with open(filename, "rb") as file:
        file.seek(fmt.offset)
        print_lzma_b85(file.read(fmt.size))

################################################################################

def get_executable_file_info(hmod):
    ids = kernel.get_resource_names(hmod, RT_GROUP_ICON)
    res = []
    for i in ids:
        addr, _ = kernel.get_resource_info(hmod, i, RT_GROUP_ICON)
        idh = ICONDIR_HEADER.from_address(addr)
        addr += ctypes.sizeof(ICONDIR_HEADER)
        ICON_ENTRIES = GRPICONDIRENTRY * idh.count
        entries = ICON_ENTRIES.from_address(addr)
        image_info = [
            SimpleNamespace(
                width=e.width if e.width else 256,
                height=e.height if e.height else 256,
                colors=e.colors if e.colors else 1 << min(e.bitcount, 24),
                ico_id=e.ico_id,
                )
            for e in entries
            ]
        res.append((i, image_info))
    return res

################################################################################

def extract_from_executable(hmod, fmt):
    addr, size = kernel.get_resource_info(hmod, fmt.ico_id, RT_ICON)
    print_lzma_b85(ctypes.string_at(addr, size))

################################################################################

def print_info(infos):
    indent = "    " if len(infos) > 1 else ""
    for iid, fmts in infos:
        if indent:
            print(f"icon id {iid}:")
        for n, fmt in enumerate(fmts):
            w = fmt.width
            h = fmt.height
            c = fmt.colors
            c = str(c) if c < 1_000_000 else f"{c >> 20}M"
            print(f"{indent}format {n:2}: {w}.{h}.{c}")
    print()

################################################################################

def main():
    if len(sys.argv) < 2:
        print("need file name (icon or executable)", file=sys.stderr)
        return 1
    filename = sys.argv[1]

    # ============== gather info =================

    flags = LOAD_LIBRARY_AS_IMAGE_RESOURCE | LOAD_LIBRARY_AS_DATAFILE
    try:
        hmod = kernel.LoadLibraryEx(filename, flags)
    except OSError:
        hmod = None

    if hmod is None:
        infos = get_ico_file_info(filename)
    else:
        infos = get_executable_file_info(hmod)

    # === print info if no further arguments =====

    if len(sys.argv) == 2:
        print_info(infos)
        return 0

    # ======== process further arguments =========

    if len(infos) > 1:
        if len(sys.argv) == 3:
            print("need icon AND format ID", file=sys.stderr)
            return 2
        try:
            ico_id = int(sys.argv[2])
        except ValueError:
            # also allow string identifier
            ico_id = sys.argv[2]
        fmt_id = int(sys.argv[3])
    else:
        ico_id = infos[0][0]
        fmt_id = int(sys.argv[2])

    # ============= try to extract ===============

    for iid, fmts in infos:
        if iid == ico_id:
            fmt = fmts[fmt_id]
            break
    else:
        raise ValueError("invalid icon id")

    if hmod is None:
        extract_from_ico(filename, fmt)
    else:
        extract_from_executable(hmod, fmt)

    return 0

################################################################################

if __name__ == "__main__":
    sys.exit(main())

################################################################################
